# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""The log viewer implementation"""

from utils.pixmapcache import getIcon
from utils.colorfont import getZoomedMonoFont
from .qt import (Qt, QSize, QColor, QBrush, QTextCursor, QCursor,
                 QPlainTextEdit, QMenu, QHBoxLayout, QWidget, QAction,
                 QToolBar)
from .spacers import ToolBarExpandingSpacer

MAX_LINES = 10000


class LogViewer(QWidget):

    """The log (+stdout, +stderr) viewer widget"""

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        self.__isEmpty = True
        self.__copyAvailable = False
        self.clearButton = None
        self.messages = None
        self.copyButton = None
        self.selectAllButton = None
        self.__createLayout(parent)
        self.onTextZoomChanged()

        # create the context menu
        self.__menu = QMenu(self)
        self.__selectAllMenuItem = self.__menu.addAction(
            getIcon('selectall.png'), 'Select All', self.messages.selectAll)
        self.__copyMenuItem = self.__menu.addAction(
            getIcon('copymenu.png'), 'Copy', self.messages.copy)
        self.__menu.addSeparator()
        self.__clearMenuItem = self.__menu.addAction(
            getIcon('trash.png'), 'Clear', self.__clear)

        self.messages.setContextMenuPolicy(Qt.CustomContextMenu)
        self.messages.customContextMenuRequested.connect(
            self.__handleShowContextMenu)
        self.messages.copyAvailable.connect(self.__onCopyAvailable)

        self.cNormalFormat = self.messages.currentCharFormat()
        self.cErrorFormat = self.messages.currentCharFormat()
        self.cErrorFormat.setForeground(QBrush(QColor(Qt.red)))
        self.__updateToolbarButtons()

    def onTextZoomChanged(self):
        """Triggered when a mono font is chosen or zoom changed"""
        self.messages.setFont(getZoomedMonoFont())

    def __createLayout(self, parent):
        """Helper to create the viewer layout"""
        # Messages list area
        self.messages = QPlainTextEdit(parent)
        self.messages.setLineWrapMode(QPlainTextEdit.NoWrap)
        self.messages.setReadOnly(True)
        self.messages.setMaximumBlockCount(MAX_LINES)

        # Default font size is good enough for most of the systems.
        # 12.0 might be good only in case of the XServer on PC (Xming).
        # self.messages.setFontPointSize( 12.0 )

        # Buttons
        self.selectAllButton = QAction(getIcon('selectall.png'),
                                       'Select all', self)
        self.selectAllButton.triggered.connect(self.messages.selectAll)
        self.copyButton = QAction(getIcon('copymenu.png'),
                                  'Copy to clipboard', self)
        self.copyButton.triggered.connect(self.messages.copy)
        self.clearButton = QAction(getIcon('trash.png'), 'Clear all', self)
        self.clearButton.triggered.connect(self.__clear)

        # Toolbar
        self.toolbar = QToolBar()
        self.toolbar.setOrientation(Qt.Vertical)
        self.toolbar.setMovable(False)
        self.toolbar.setAllowedAreas(Qt.LeftToolBarArea)
        self.toolbar.setIconSize(QSize(16, 16))
        self.toolbar.setFixedWidth(28)
        self.toolbar.setContentsMargins(0, 0, 0, 0)
        self.toolbar.addAction(self.selectAllButton)
        self.toolbar.addAction(self.copyButton)
        self.toolbar.addWidget(ToolBarExpandingSpacer(self.toolbar))
        self.toolbar.addAction(self.clearButton)

        # layout
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.toolbar)
        layout.addWidget(self.messages)

        self.setLayout(layout)

    # Arguments: coord
    def __handleShowContextMenu(self, _):
        """Show the context menu"""
        self.__selectAllMenuItem.setEnabled(not self.__isEmpty)
        self.__copyMenuItem.setEnabled(self.__copyAvailable)
        self.__clearMenuItem.setEnabled(not self.__isEmpty)

        self.__menu.popup(QCursor.pos())

    def __appendText(self, txt, isError):
        """Append the text"""
        if txt:
            self.__isEmpty = False
            cursor = self.messages.textCursor()
            cursor.movePosition(QTextCursor.End)
            self.messages.setTextCursor(cursor)
            if isError:
                self.messages.setCurrentCharFormat(self.cErrorFormat)
            else:
                self.messages.setCurrentCharFormat(self.cNormalFormat)
            self.messages.insertPlainText(txt)
            self.messages.insertPlainText('\n')
            self.messages.ensureCursorVisible()
            self.__updateToolbarButtons()

    def appendMessage(self, txt):
        """Append the regular message"""
        self.__appendText(txt, False)

    def appendError(self, txt):
        """Append the error message"""
        self.__appendText(txt, True)

    def append(self, txt):
        """Decides what the message is - error or not - and append it then"""
        if txt.startswith('CRITICAL') or \
           txt.startswith('ERROR') or \
           txt.startswith('WARNING'):
            self.appendError(txt)
        else:
            self.appendMessage(txt)

    def __updateToolbarButtons(self):
        """Contextually updates toolbar buttons"""
        self.selectAllButton.setEnabled(not self.__isEmpty)
        self.copyButton.setEnabled(self.__copyAvailable)
        self.clearButton.setEnabled(not self.__isEmpty)

    def __clear(self):
        """Triggers when the clear function is selected"""
        self.__isEmpty = True
        self.__copyAvailable = False
        self.messages.clear()
        self.__updateToolbarButtons()

    def __onCopyAvailable(self, isAvailable):
        """Triggers on the copyAvailable signal"""
        self.__copyAvailable = isAvailable
        self.__updateToolbarButtons()

    def getText(self):
        """Provides the content as a plain text"""
        return self.messages.toPlainText()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017 Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


"""Breakpoints margin: shows breakpoints and the current debug line"""

import os.path
import logging
import math
from sys import maxsize
from qutepart.margins import MarginBase
import qutepart
from ui.qt import QWidget, QPainter, QModelIndex, QToolTip
from utils.misc import extendInstance
from utils.pixmapcache import getPixmap
from utils.fileutils import isPythonMime
from utils.globals import GlobalData
from utils.settings import Settings
from debugger.bputils import getBreakpointLines
from debugger.breakpoint import Breakpoint


# Note: it incorporates all the editor related breakpoints functionality:
#       - creating a new one via a click
#       - updating the line numbers if needed
#       - validating
#       - etc.


def getMarginBits():
    """Provides the number of block value bits to cover maxBreakpoints value"""
    # 0 means no block value so +1 is required
    distinctValues = Settings()['maxBreakpoints'] + 1
    bits = 1
    while True:
        if 2**bits >= distinctValues:
            return bits
        bits += 1


class CDMBreakpointMargin(QWidget):

    """Breakpoints area widget"""

    BPOINT_MARK = 1
    TMP_BPOINT_MARK = 2
    DISABLED_BPOINT_MARK = 3

    def __init__(self, parent, debugger):
        QWidget.__init__(self, parent)

        extendInstance(self, MarginBase)
        MarginBase.__init__(self, parent,
                            "cdm_bpoint_margin", getMarginBits())
        self.setMouseTracking(True)

        self.__debugger = debugger
        self.__breakpoints = {}     # block handle -> Breakpoint instance
        self.__breakableLines = None
        self.__maxBreakpoints = Settings()['maxBreakpoints']
        self.__bgColor = GlobalData().skin['bpointsMarginPaper']

        self.__marks = {
            self.BPOINT_MARK: [getPixmap('dbgbpointmarker.png'), 0],
            self.TMP_BPOINT_MARK: [getPixmap('dbgtmpbpointmarker.png'), 0],
            self.DISABLED_BPOINT_MARK: [getPixmap('dbgdisbpointmarker.png'), 0]}

        for item in self.__marks:
            self.__marks[item][1] = self.__marks[item][0].height()
            if self.__marks[item][0].height() != self.__marks[item][0].width():
                logging.error('breakpoint margin pixmap needs to be square')

        self.myUUID = None
        if hasattr(self._qpart._parent, 'getUUID'):
            self.myUUID = self._qpart._parent.getUUID()

        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager
        editorsManager.sigFileTypeChanged.connect(self.__onFileTypeChanged)

        self.blockClicked.connect(self.__onBlockClicked)

        self._qpart.blockCountChanged.connect(self.__onBlockCountChanged)

        bpointModel = self.__debugger.getBreakPointModel()
        bpointModel.rowsAboutToBeRemoved.connect(self.__deleteBreakPoints)
        bpointModel.sigDataAboutToBeChanged.connect(
            self.__breakPointDataAboutToBeChanged)
        bpointModel.dataChanged.connect(self.__changeBreakPoints)
        bpointModel.rowsInserted.connect(self.__addBreakPoints)

    def paintEvent(self, event):
        """Paint the margin"""
        painter = QPainter(self)
        painter.fillRect(event.rect(), self.__bgColor)
        oneLineHeight = self._qpart.fontMetrics().height()

        block = self._qpart.firstVisibleBlock()
        geometry = self._qpart.blockBoundingGeometry(block)
        blockBoundingGeometry = geometry.translated(
            self._qpart.contentOffset())
        top = blockBoundingGeometry.top()
        # bottom = top + blockBoundingGeometry.height()

        for block in qutepart.iterateBlocksFrom(block):
            height = self._qpart.blockBoundingGeometry(block).height()
            if top > event.rect().bottom():
                break
            if block.isVisible():
                # lineNo = block.blockNumber() + 1
                blockValue = self.getBlockValue(block)
                pixmap = None
                if blockValue != 0:
                    bpoint = self.__breakpoints[blockValue]
                    if not bpoint.isEnabled():
                        markType = self.DISABLED_BPOINT_MARK
                    elif bpoint.isTemporary():
                        markType = self.TMP_BPOINT_MARK
                    else:
                        markType = self.BPOINT_MARK
                    pixmap, edge = self.__marks[markType]

                if pixmap:
                    xPos = 0
                    yPos = top
                    if edge <= oneLineHeight:
                        yPos += math.ceil((oneLineHeight - edge) / 2)
                    else:
                        edge = oneLineHeight
                        xPos = math.ceil((self.width() - edge) / 2)
                    painter.drawPixmap(xPos, yPos, edge, edge, pixmap)
            top += height

    def mouseMoveEvent(self, event):
        """Tooltips for the marks"""
        if self.__breakpoints:
            textCursor = self._qpart.cursorForPosition(event.pos())
            block = textCursor.block()
            msg = None

            blockValue = self.getBlockValue(block)
            if blockValue != 0:
                bpoint = self.__breakpoints[blockValue]
                if not bpoint.isEnabled():
                    msg = 'Disabled breakpoint'
                elif bpoint.isTemporary():
                    msg = 'Temporary breakpoint'
                else:
                    msg = 'Regular breakpoint'

            if msg:
                QToolTip.showText(event.globalPos(), msg)
            else:
                QToolTip.hideText()
        return QWidget.mouseMoveEvent(self, event)

    @staticmethod
    def width():
        """Desired width"""
        return 16

    def setBackgroundColor(self, color):
        """Sets the new background color"""
        if self.__bgColor != color:
            self.__bgColor = color
            self.update()

    def __onFileTypeChanged(self, fileName, uuid, newFileType):
        """Triggered on the changed file type"""
        del fileName    # unused argument
        if uuid == self.myUUID:
            MarginBase.setVisible(self, isPythonMime(newFileType))

    def restoreBreakpoints(self):
        """Restores the breakpoints"""
        for _, bpoint in self.__breakpoints.items():
            line = bpoint.getLineNumber()
            self.setBlockValue(
                self._qpart.document().findBlockByNumber(line - 1), 0)
        self.__breakpoints = {}
        self.__addBreakPoints(
            QModelIndex(), 0,
            self.__debugger.getBreakPointModel().rowCount() - 1)
        self.validateBreakpoints()
        self.update()

    def setDebugMode(self, debugOn, disableEditing):
        """Called to switch between debug/development"""
        del debugOn         # unused argument
        del disableEditing  # unused argument
        self.__breakableLines = None

    def isLineBreakable(self, line=None, enforceRecalc=False,
                        enforceSure=False):
        """True if a breakpoint could be placed on the current line"""
        fileName = self._qpart._parent.getFileName()

        if fileName is None or fileName == "" or not os.path.isabs(fileName):
            return False
        if not self._qpart.isPythonBuffer():
            return False

        if line is None:
            curPos = self._qpart.cursorPosition
            line = curPos[0] + 1
        if self.__breakableLines is not None and not enforceRecalc:
            return line in self.__breakableLines

        self.__breakableLines = getBreakpointLines(fileName, self._qpart.text,
                                                   enforceRecalc)

        if self.__breakableLines is None:
            if not enforceSure:
                # Be on the safe side - if there is a problem of
                # getting the breakable lines, let the user decide
                return True
            return False

        return line in self.__breakableLines

    def newBreakpointWithProperties(self, bpoint):
        """Sets a new breakpoint and its properties"""
        if len(self.__breakpoints) >= Settings()['maxBreakpoints']:
            logging.error('The max number of breakpoints per file (' +
                          str(Settings()['maxBreakpoints']) +
                          ') is exceeded')
            return

        line = bpoint.getLineNumber()
        bpointHandle = self.__getAvailableHandle()
        self.__breakpoints[bpointHandle] = bpoint
        self.setBlockValue(self._qpart.document().findBlockByNumber(line - 1),
                           bpointHandle)
        self.update()

    def deleteAllBreakpoints(self):
        """Deletes all the breakpoints in the buffer"""
        self.__deleteBreakPointsInLineRange(1, self._qpart.lines())

    def clearBreakpoint(self, line):
        """Clears a breakpoint"""
        for handle, bpoint in self.__breakpoints.items():
            if bpoint.getLineNumber() == line:
                self.setBlockValue(
                    self._qpart.document().findBlockByNumber(line - 1), 0)
                del self.__breakpoints[handle]
                self.update()
                return

    def validateBreakpoints(self):
        """Checks breakpoints and deletes those which are invalid"""
        if not self.__breakpoints:
            return

        fileName = self._qpart._parent.getFileName()
        breakableLines = getBreakpointLines(fileName, self._qpart.text,
                                            True, False)

        toBeDeleted = []
        for _, bpoint in self.__breakpoints.items():
            bpointLine = bpoint.getLineNumber()
            if breakableLines is None or bpointLine not in breakableLines:
                toBeDeleted.append(bpointLine)

        if toBeDeleted:
            model = self.__debugger.getBreakPointModel()
            for line in toBeDeleted:
                location = ':'.join([fileName, str(line)])
                if breakableLines is None:
                    msg = 'Breakpoint at ' + location + ' does not point to ' \
                          'a breakable line (file is not compilable).'
                else:
                    msg = 'Breakpoint at ' + location + ' does not point to ' \
                          'a breakable line anymore.'
                logging.warning(msg + ' The breakpoint is deleted.')
                index = model.getBreakPointIndex(fileName, line)
                self.setBlockValue(
                    self._qpart.document().findBlockByNumber(line - 1), 0)
                model.deleteBreakPointByIndex(index)

    def __getAvailableHandle(self):
        """Provides the available handle"""
        occupiedHandles = list(self.__breakpoints.keys())
        vacantHandle = 1
        while True:
            if vacantHandle not in occupiedHandles:
                return vacantHandle
            vacantHandle += 1

    def __addBreakpoint(self, line, temporary, enabled=True):
        """Adds a new breakpoint"""
        # The prerequisites:
        # - it is saved buffer
        # - it is a python buffer
        # - it is a breakable line
        # are checked in the function
        if not self.isLineBreakable(line, True, True):
            return

        fileName = self._qpart._parent.getFileName()
        bpoint = Breakpoint(fileName, line, "", temporary, enabled, 0)
        self.__debugger.getBreakPointModel().addBreakpoint(bpoint)

    def __addBreakPoints(self, parentIndex, start, end):
        """Adds breakpoints"""
        bpointModel = self.__debugger.getBreakPointModel()
        fileName = self._qpart._parent.getFileName()

        for row in range(start, end + 1):
            index = bpointModel.index(row, 0, parentIndex)
            bpoint = bpointModel.getBreakPointByIndex(index)
            bpFileName = bpoint.getAbsoluteFileName()

            if bpFileName == fileName:
                self.newBreakpointWithProperties(bpoint)

    def __deleteBreakPoints(self, parentIndex, start, end):
        """Deletes breakpoints"""
        bpointModel = self.__debugger.getBreakPointModel()
        fileName = self._qpart._parent.getFileName()

        for row in range(start, end + 1):
            index = bpointModel.index(row, 0, parentIndex)

            bpoint = bpointModel.getBreakPointByIndex(index)
            bpFileName = bpoint.getAbsoluteFileName()

            if bpFileName == fileName:
                self.clearBreakpoint(bpoint.getLineNumber())

    def __changeBreakPoints(self, startIndex, endIndex):
        """Sets changed breakpoints"""
        self.__addBreakPoints(QModelIndex(),
                              startIndex.row(), endIndex.row())

    def __breakPointDataAboutToBeChanged(self, startIndex, endIndex):
        """Handles the dataAboutToBeChanged signal of the breakpoint model"""
        self.__deleteBreakPoints(QModelIndex(),
                                 startIndex.row(), endIndex.row())

    def __toggleBreakpoint(self, line, temporary=False):
        """Toggles the line breakpoint"""
        # Clicking loop: none->regular->temporary->disabled->none
        fileName = self._qpart._parent.getFileName()
        model = self.__debugger.getBreakPointModel()
        for _, bpoint in self.__breakpoints.items():
            if bpoint.getLineNumber() == line:
                index = model.getBreakPointIndex(fileName, line)
                model.deleteBreakPointByIndex(index)
                if not bpoint.isEnabled():
                    self.setBlockValue(
                        self._qpart.document().findBlockByNumber(line - 1), 0)
                elif bpoint.isTemporary():
                    self.__addBreakpoint(line, False, False)
                else:
                    self.__addBreakpoint(line, True)
                return
        if len(self.__breakpoints) < self.__maxBreakpoints:
            self.__addBreakpoint(line, temporary)
        else:
            logging.error('Max breakpoint number per file (' +
                          str(self.__maxBreakpoints) +
                          ') has been reached. The breakpoint is not added.')

    def __deleteBreakPointsInLineRange(self, startFrom, count):
        """Deletes breakpoints which fall into the given lines range"""
        toBeDeleted = []
        limit = startFrom + count - 1
        for _, bpoint in self.__breakpoints.items():
            bpointLine = bpoint.getLineNumber()
            if bpointLine >= startFrom and bpointLine <= limit:
                toBeDeleted.append(bpointLine)

        if toBeDeleted:
            model = self.__debugger.getBreakPointModel()
            fileName = self._qpart._parent.getFileName()
            for line in toBeDeleted:
                index = model.getBreakPointIndex(fileName, line)
                self.setBlockValue(
                    self._qpart.document().findBlockByNumber(line - 1), 0)
                model.deleteBreakPointByIndex(index)

    def __onBlockClicked(self, block):
        """Margin of the block has been clicked"""
        lineNo = block.blockNumber() + 1
        for _, bpoint in self.__breakpoints.items():
            if bpoint.getLineNumber() == lineNo:
                # Breakpoint marker is set for this line already
                self.__toggleBreakpoint(lineNo)
                return

        # Check if it is a python file
        if not self._qpart.isPythonBuffer():
            return

        fileName = self._qpart._parent.getFileName()
        if fileName is None or fileName == "" or not os.path.isabs(fileName):
            logging.warning("The buffer has to be saved "
                            "before setting breakpoints")
            return

        breakableLines = getBreakpointLines("", self._qpart.text, True, False)
        if breakableLines is None:
            logging.warning("The breakable lines could not be identified "
                            "due to the file compilation errors. Fix the code "
                            "and try again.")
            return

        breakableLines = list(breakableLines)
        breakableLines.sort()
        if not breakableLines:
            logging.warning("There are no breakable lines")
            return

        if lineNo in breakableLines:
            self.__toggleBreakpoint(lineNo)
            return

        # There are breakable lines however the user requested a line which
        # is not breakable
        candidateLine = breakableLines[0]
        if lineNo < breakableLines[0]:
            candidateLine = breakableLines[0]
        elif lineNo > breakableLines[-1]:
            candidateLine = breakableLines[-1]
        else:
            lowerDistance = maxsize
            upperDistance = maxsize

            for breakableLine in breakableLines:
                if breakableLine < lineNo:
                    lowerDistance = min(lowerDistance, lineNo - breakableLine)
                else:
                    upperDistance = min(upperDistance, breakableLine - lineNo)
            if lowerDistance < upperDistance:
                candidateLine = lineNo - lowerDistance
            else:
                candidateLine = lineNo + upperDistance

        if not self._qpart.isLineOnScreen(candidateLine - 1):
            # The redirected breakpoint line is not on the screen, scroll it
            self._qpart.ensureLineOnScreen(candidateLine - 1)
            self._qpart.setFirstVisible(max(0, candidateLine - 2))

        self.__toggleBreakpoint(candidateLine)

    def __onBlockCountChanged(self):
        """Number of lines in the file has changed"""
        if not self.__breakpoints:
            return

        oldSet = set(self.__breakpoints.keys())

        currentSet = set()
        currentHandleToLine = {}
        startBlock = self._qpart.document().firstBlock()
        for block in qutepart.iterateBlocksFrom(startBlock):
            handle = self.getBlockValue(block)
            if handle != 0:
                currentSet.add(handle)
                currentHandleToLine[handle] = block.blockNumber() + 1

        deletedHandles = oldSet - currentSet

        changedLineHandles = set()
        for cHandle, cLine in currentHandleToLine.items():
            if self.__breakpoints[cHandle].getLineNumber() != cLine:
                changedLineHandles.add(cHandle)

        if deletedHandles or changedLineHandles:
            fileName = self._qpart._parent.getFileName()
            model = self.__debugger.getBreakPointModel()

            for deletedHandle in deletedHandles:
                deletedBP = self.__breakpoints[deletedHandle]
                deletedLine = deletedBP.getLineNumber()
                del self.__breakpoints[deletedHandle]
                index = model.getBreakPointIndex(fileName, deletedLine)
                model.deleteBreakPointByIndex(index)

            for changedHandle in changedLineHandles:
                newLineNo = currentHandleToLine[changedHandle]
                self.__breakpoints[changedHandle].updateLineNumber(newLineNo)
                index = model.getBreakPointIndex(fileName, newLineNo)
                model.updateLineNumberByIndex(index, newLineNo)

    def onClose(self):
        """The editor is going to be closed"""
        # Prevent clearing breakpoints when an editor is closed.
        # The signal is generated when the text is deleted and it happens when
        # the editor is destroyed
        self._qpart.blockCountChanged.disconnect(self.__onBlockCountChanged)

        # Release the text editor link
        self._qpart = None

        # Editors manager dependency
        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager
        editorsManager.sigFileTypeChanged.disconnect(self.__onFileTypeChanged)

        self.blockClicked.disconnect(self.__onBlockClicked)

        bpointModel = self.__debugger.getBreakPointModel()
        bpointModel.rowsAboutToBeRemoved.disconnect(self.__deleteBreakPoints)
        bpointModel.sigDataAboutToBeChanged.disconnect(
            self.__breakPointDataAboutToBeChanged)
        bpointModel.dataChanged.disconnect(self.__changeBreakPoints)
        bpointModel.rowsInserted.disconnect(self.__addBreakPoints)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Navigation bar implementation"""

from ui.qt import (QTimer, Qt, QEvent, pyqtSignal, QFrame, QHBoxLayout, QLabel,
                   QWidget, QSizePolicy, QComboBox)
from utils.globals import GlobalData
from utils.settings import Settings
from utils.fileutils import isPythonMime
from utils.pixmapcache import getPixmap, getIcon
from cdmpyparser import getBriefModuleInfoFromMemory
from autocomplete.bufferutils import getContext


IDLE_TIMEOUT = 1500


class NavBarComboBox(QComboBox):

    """Navigation bar combo box"""

    jumpToLine = pyqtSignal(int)

    def __init__(self, parent=None):
        QComboBox.__init__(self, parent)
        self.setSizeAdjustPolicy(QComboBox.AdjustToMinimumContentsLength)
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHeightForWidth(self.sizePolicy().hasHeightForWidth())
        self.setSizePolicy(sizePolicy)
        self.activated.connect(self.onActivated)

        self.view().installEventFilter(self)

        self.pathIndex = None

    # Arguments: obj, event
    def eventFilter(self, _, event):
        """Event filter for the qcombobox list view"""
        if event.type() == QEvent.KeyPress:
            key = event.key()
            if key == Qt.Key_Escape:
                self.parent().getEditor().setFocus()
                return True
            if key == Qt.Key_Left:
                # Move cursor to the left combo
                if self.pathIndex is not None:
                    self.parent().activateCombo(self, self.pathIndex - 1)
                    return True
            if key == Qt.Key_Right:
                # Move cursor to the right combo
                if self.pathIndex is not None:
                    self.parent().activateCombo(self, self.pathIndex + 1)
                else:
                    self.parent().activateCombo(self, 0)
                return True
        return False

    def onActivated(self, index):
        """User selected an item"""
        if index >= 0:
            self.jumpToLine.emit(self.itemData(index))


class PathElement:

    """Single path element"""

    def __init__(self, parent=None):
        self.icon = QLabel()
        self.icon.setPixmap(getPixmap('nbsep.png'))
        self.combo = NavBarComboBox(parent)


class NavigationBar(QFrame):

    """Navigation bar at the top of the editor (python only)"""

    STATE_OK_UTD = 0        # Parsed OK, context up to date
    STATE_OK_CHN = 1        # Parsed OK, context changed
    STATE_BROKEN_UTD = 2    # Parsed with errors, context up to date
    STATE_BROKEN_CHN = 3    # Parsed with errors, context changed
    STATE_UNKNOWN = 4

    def __init__(self, editor, parent):
        QFrame.__init__(self, parent)
        self.__editor = editor
        self.__parentWidget = parent

        # It is always not visible at the beginning because there is no
        # editor content at the start
        self.setVisible(False)

        # There is no parser info used to display values
        self.__currentInfo = None
        self.__currentIconState = self.STATE_UNKNOWN
        self.__connected = False

        # List of PathElement starting after the global scope
        self.__path = []

        self.__createLayout()

        # Create the update timer
        self.__updateTimer = QTimer(self)
        self.__updateTimer.setSingleShot(True)
        self.__updateTimer.timeout.connect(self.updateBar)

        # Connect to the change file type signal
        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager
        editorsManager.sigFileTypeChanged.connect(self.__onFileTypeChanged)

    def terminate(self):
        """Closing up"""
        self.__disconnectEditorSignals()
        self.__updateTimer.stop()
        self.__updateTimer.timeout.disconnect(self.updateBar)

        # Editors manager dependency
        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager
        editorsManager.sigFileTypeChanged.disconnect(self.__onFileTypeChanged)

        # Combo boxes signals
        self.__globalScopeCombo.jumpToLine.disconnect(self.__onJumpToLine)
        for item in self.__path:
            item.combo.jumpToLine.disconnect(self.__onJumpToLine)

        # Release the editor reference
        self.__editor = None

    def getEditor(self):
        """Provides the editor"""
        return self.__editor

    def __connectEditorSignals(self):
        """When it is a python file - connect to the editor signals"""
        if not self.__connected:
            self.__editor.cursorPositionChanged.connect(
                self.__cursorPositionChanged)
            self.__editor.textChanged.connect(self.__onBufferChanged)
            self.__connected = True

    def __disconnectEditorSignals(self):
        """Disconnect the editor signals when the file is not a python one"""
        if self.__connected:
            self.__editor.cursorPositionChanged.disconnect(
                self.__cursorPositionChanged)
            self.__editor.textChanged.disconnect(self.__onBufferChanged)
            self.__connected = False

    def __createLayout(self):
        """Creates the layout"""
        self.__layout = QHBoxLayout(self)
        self.__layout.setContentsMargins(0, 0, 0, 0)

        # Set the background color

        # Create info icon
        self.__infoIcon = QLabel()
        self.__layout.addWidget(self.__infoIcon)

        self.__globalScopeCombo = NavBarComboBox(self)
        self.__globalScopeCombo.jumpToLine.connect(self.__onJumpToLine)
        self.__layout.addWidget(self.__globalScopeCombo)

        self.__spacer = QWidget()
        self.__spacer.setSizePolicy(QSizePolicy.Expanding,
                                    QSizePolicy.Expanding)
        self.__layout.addWidget(self.__spacer)
        self.setFixedHeight(self.__globalScopeCombo.height())

    def __updateInfoIcon(self, state):
        """Updates the information icon"""
        if state != self.__currentIconState:
            if state == self.STATE_OK_UTD:
                self.__infoIcon.setPixmap(getPixmap('nbokutd.png'))
                self.__infoIcon.setToolTip("Context is up to date")
                self.__currentIconState = self.STATE_OK_UTD
            elif state == self.STATE_OK_CHN:
                self.__infoIcon.setPixmap(getPixmap('nbokchn.png'))
                self.__infoIcon.setToolTip("Context is not up to date; "
                                           "will be updated on idle")
                self.__currentIconState = self.STATE_OK_CHN
            elif state == self.STATE_BROKEN_UTD:
                self.__infoIcon.setPixmap(getPixmap('nbbrokenutd.png'))
                self.__infoIcon.setToolTip("Context might be invalid "
                                           "due to invalid python code")
                self.__currentIconState = self.STATE_BROKEN_UTD
            else:
                # STATE_BROKEN_CHN
                self.__infoIcon.setPixmap(getPixmap('nbbrokenchn.png'))
                self.__infoIcon.setToolTip("Context might be invalid; "
                                           "will be updated on idle")
                self.__currentIconState = self.STATE_BROKEN_CHN

    def resizeEvent(self, event):
        """Editor has resized"""
        QFrame.resizeEvent(self, event)

    # Arguments: fileName, uuid, newFileType
    def __onFileTypeChanged(self, _, uuid, newFileType):
        """Triggered when a buffer content type has changed"""
        if self.__parentWidget.getUUID() != uuid:
            return

        if isPythonMime(newFileType) and Settings()['showNavigationBar']:
            # Update the bar and show it
            self.setVisible(True)
            self.updateBar()
        else:
            self.__disconnectEditorSignals()
            self.__updateTimer.stop()
            self.__currentInfo = None
            self.setVisible(False)
            self.__currentIconState = self.STATE_UNKNOWN

    def updateSettings(self):
        """Called when navigation bar settings have been updated"""
        textMime = self.__parentWidget.getMime()
        if Settings()['showNavigationBar'] and isPythonMime(textMime):
            self.setVisible(True)
            self.updateBar()
        else:
            self.__disconnectEditorSignals()
            self.__updateTimer.stop()
            self.__currentInfo = None
            self.setVisible(False)

    def updateBar(self):
        """Triggered when the timer is fired"""
        self.__updateTimer.stop()  # just in case

        if not isPythonMime(self.__parentWidget.getMime()):
            return

        if not self.__connected:
            self.__connectEditorSignals()

        # Parse the buffer content
        self.__currentInfo = getBriefModuleInfoFromMemory(self.__editor.text)

        # Decide what icon to use
        if self.__currentInfo.isOK:
            self.__updateInfoIcon(self.STATE_OK_UTD)
        else:
            self.__updateInfoIcon(self.STATE_BROKEN_UTD)

        # Calc the cursor context
        context = getContext(self.__editor, self.__currentInfo, True, False)

        # Display the context
        self.__populateGlobalScope()
        if context.length == 0:
            self.__globalScopeCombo.setCurrentIndex(-1)
        else:
            index = self.__globalScopeCombo.findData(
                context.levels[0][0].line)
            self.__globalScopeCombo.setCurrentIndex(index)

        usedFromStore = 0
        index = 1
        while index < context.length:
            if len(self.__path) < index:
                newPathItem = PathElement(self)
                self.__path.append(newPathItem)
                self.__layout.addWidget(newPathItem.icon)
                self.__layout.addWidget(newPathItem.combo)
                combo = newPathItem.combo
                combo.pathIndex = len(self.__path) - 1
                combo.jumpToLine.connect(self.__onJumpToLine)
            else:
                self.__path[index - 1].icon.setVisible(True)
                self.__path[index - 1].combo.setVisible(True)
                combo = self.__path[index - 1].combo
                combo.clear()

            # Populate the combo box
            self.__populateClassesAndFunctions(context.levels[index - 1][0],
                                               combo)
            combo.setCurrentIndex(combo.findData(context.levels[index][0].line))
            index += 1
            usedFromStore += 1

        # it might need to have one more level with nothing selected
        if context.length > 0:
            if len(context.levels[context.length - 1][0].functions) > 0 or \
               len(context.levels[context.length - 1][0].classes) > 0:
                # Need to add a combo
                if len(self.__path) <= usedFromStore:
                    newPathItem = PathElement(self)
                    self.__path.append(newPathItem)
                    self.__layout.addWidget(newPathItem.icon)
                    self.__layout.addWidget(newPathItem.combo)
                    combo = newPathItem.combo
                    combo.pathIndex = len(self.__path) - 1
                    combo.jumpToLine.connect(self.__onJumpToLine)
                else:
                    self.__path[index - 1].icon.setVisible(True)
                    self.__path[index - 1].combo.setVisible(True)
                    combo = self.__path[index - 1].combo
                    combo.clear()

                self.__populateClassesAndFunctions(
                    context.levels[context.length - 1][0], combo)
                combo.setCurrentIndex(-1)
                usedFromStore += 1

        # Hide extra components if so
        index = usedFromStore
        while index < len(self.__path):
            self.__path[index].icon.setVisible(False)
            self.__path[index].combo.setVisible(False)
            index += 1

        # Make sure the spacer is the last item
        self.__layout.removeWidget(self.__spacer)
        self.__layout.addWidget(self.__spacer)

    def __populateGlobalScope(self):
        """Repopulates the global scope combo box"""
        self.__globalScopeCombo.clear()

        self.__populateClassesAndFunctions(self.__currentInfo,
                                           self.__globalScopeCombo)

        if not Settings()['navbarglobalsimports']:
            return

        if len(self.__currentInfo.globals) == 0 and \
           len(self.__currentInfo.imports) == 0:
            return

        if self.__globalScopeCombo.count() != 0:
            self.__globalScopeCombo.insertSeparator(
                self.__globalScopeCombo.count())

        for glob in self.__currentInfo.globals:
            self.__globalScopeCombo.addItem('global ' + glob.name, glob.line)
        for imp in self.__currentInfo.imports:
            self.__globalScopeCombo.addItem('import ' + imp.name, imp.line)

    @staticmethod
    def __populateClassesAndFunctions(infoObj, combo):
        """Populates the combo with classes and functions from the infoObj"""
        for klass in infoObj.classes:
            combo.addItem('class ' + klass.name, klass.line)
        for func in infoObj.functions:
            combo.addItem('def ' + func.name, func.line)

    def __cursorPositionChanged(self):
        """Cursor position changed"""
        self.__onNeedUpdate()

    def __onBufferChanged(self):
        """Buffer changed"""
        self.__onNeedUpdate()

    def __onNeedUpdate(self):
        """Triggered to update status icon and to restart the timer"""
        self.__updateTimer.stop()
        if self.__currentInfo.isOK:
            self.__updateInfoIcon(self.STATE_OK_CHN)
        else:
            self.__updateInfoIcon(self.STATE_BROKEN_CHN)
        self.__updateTimer.start(IDLE_TIMEOUT)

    def __onJumpToLine(self, line):
        """Triggered when it needs to jump to a line"""
        self.__editor.gotoLine(line, 0)
        self.__editor.setFocus()

    def setFocusToLastCombo(self):
        """Activates the last combo"""
        if self.__currentInfo is None:
            return
        for index in range(len(self.__path) - 1, -1, -1):
            if self.__path[index].combo.isVisible():
                self.__path[index].combo.setFocus()
                self.__path[index].combo.showPopup()
                return

        self.__globalScopeCombo.setFocus()
        self.__globalScopeCombo.showPopup()

    def activateCombo(self, currentCombo, newIndex):
        """Triggered when a neighbour combo should be activated"""
        if newIndex == -1:
            if len(self.__path) > 0:
                if self.__path[0].combo.isVisible():
                    currentCombo.hidePopup()
            self.__globalScopeCombo.setFocus()
            self.__globalScopeCombo.showPopup()
            return

        if newIndex >= len(self.__path):
            # This is the most right one
            return

        if self.__path[newIndex].combo.isVisible():
            currentCombo.hidePopup()
            self.__path[newIndex].combo.setFocus()
            self.__path[newIndex].combo.showPopup()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2015-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Various items used to represent a control flow on a virtual canvas"""

# pylint: disable=C0305

from ui.qt import Qt, QPointF, QPen, QBrush, QGraphicsRectItem, QGraphicsItem
from .auxitems import Connector, BadgeItem
from .cml import CMLVersion, CMLsw
from .colormixin import ColorMixin
from .iconmixin import IconMixin
from .cellelement import CellElement
from .textmixin import TextMixin


class CodeBlockCell(CellElement, TextMixin, ColorMixin, QGraphicsRectItem):

    """Represents a single code block"""

    def __init__(self, ref, canvas, x, y):
        CellElement.__init__(self, ref, canvas, x, y)
        TextMixin.__init__(self)
        ColorMixin.__init__(self, ref, self.canvas.settings.codeBlockBGColor,
                            self.canvas.settings.codeBlockFGColor,
                            self.canvas.settings.codeBlockBorderColor)
        QGraphicsRectItem.__init__(self, canvas.scopeRectangle)
        self.kind = CellElement.CODE_BLOCK
        self.connector = None

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        self.setupText(self)

        vPadding = 2 * (settings.vCellPadding + settings.vTextPadding)
        self.minHeight = self.textRect.height() + vPadding
        hPadding = 2 * (settings.hCellPadding + settings.hTextPadding)
        self.minWidth = max(self.textRect.width() + hPadding,
                            settings.minWidth)
        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        # Add the connector as a separate scene item to make the selection
        # working properly
        settings = self.canvas.settings
        self.connector = Connector(self.canvas, baseX + settings.mainLine,
                                   baseY,
                                   baseX + settings.mainLine,
                                   baseY + self.height)
        scene.addItem(self.connector)

        # Setting the rectangle is important for the selection and for
        # redrawing. Thus the selection pen with must be considered too.
        penWidth = settings.selectPenWidth - 1
        self.setRect(baseX + settings.hCellPadding - penWidth,
                     baseY + settings.vCellPadding - penWidth,
                     self.minWidth - 2 * settings.hCellPadding + 2 * penWidth,
                     self.minHeight - 2 * settings.vCellPadding + 2 * penWidth)
        scene.addItem(self)

    def paint(self, painter, option, widget):
        """Draws the code block"""
        del option      # unused argument
        del widget      # unused argument

        settings = self.canvas.settings
        painter.setPen(self.getPainterPen(self.isSelected(), self.borderColor))
        painter.setBrush(QBrush(self.bgColor))

        rectWidth = self.minWidth - 2 * settings.hCellPadding
        rectHeight = self.minHeight - 2 * settings.vCellPadding

        painter.drawRect(self.baseX + settings.hCellPadding,
                         self.baseY + settings.vCellPadding,
                         rectWidth, rectHeight)

        # Draw the text in the rectangle
        pen = QPen(self.fgColor)
        painter.setFont(settings.monoFont)
        painter.setPen(pen)

        textWidth = self.textRect.width() + 2 * settings.hTextPadding
        textShift = (rectWidth - textWidth) / 2
        painter.drawText(
            self.baseX + settings.hCellPadding +
            settings.hTextPadding + textShift,
            self.baseY + settings.vCellPadding + settings.vTextPadding,
            self.textRect.width(), self.textRect.height(),
            Qt.AlignLeft, self.text)

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return 'Code block at ' + \
            CellElement.getLinesSuffix(self.getLineRange())


class ReturnCell(CellElement,
                 TextMixin, ColorMixin, IconMixin, QGraphicsRectItem):

    """Represents a single return statement"""

    def __init__(self, ref, canvas, x, y):
        CellElement.__init__(self, ref, canvas, x, y)
        TextMixin.__init__(self)
        ColorMixin.__init__(self, ref, self.canvas.settings.returnBGColor,
                            self.canvas.settings.returnFGColor,
                            self.canvas.settings.returnBorderColor)
        IconMixin.__init__(self, canvas, 'return.svg', 'return')
        QGraphicsRectItem.__init__(self)
        self.kind = CellElement.RETURN
        self.connector = None

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        customText = self.ref.getDisplayValue()
        if not customText:
            customText = 'None'
        self.setupText(self, customText=customText)

        vPadding = 2 * (settings.vCellPadding + settings.vTextPadding)
        self.minHeight = self.textRect.height() + vPadding
        self.minWidth = max(
            self.textRect.width() + 2 * settings.hCellPadding +
            settings.hTextPadding + settings.returnRectRadius +
            2 * settings.hTextPadding + self.iconItem.iconWidth(),
            settings.minWidth)

        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        # Add the connector as a separate scene item to make the selection
        # working properly
        settings = self.canvas.settings
        self.connector = Connector(self.canvas, baseX + settings.mainLine,
                                   baseY,
                                   baseX + settings.mainLine,
                                   baseY + settings.vCellPadding)

        # Setting the rectangle is important for the selection and for
        # redrawing. Thus the selection pen with must be considered too.
        penWidth = settings.selectPenWidth - 1
        self.setRect(baseX + settings.hCellPadding - penWidth,
                     baseY + settings.vCellPadding - penWidth,
                     self.minWidth - 2 * settings.hCellPadding + 2 * penWidth,
                     self.minHeight - 2 * settings.vCellPadding + 2 * penWidth)

        self.iconItem.setPos(
            baseX + settings.hCellPadding + settings.hTextPadding,
            baseY + self.minHeight/2 - self.iconItem.iconHeight()/2)

        scene.addItem(self.connector)
        scene.addItem(self)
        scene.addItem(self.iconItem)

    def paint(self, painter, option, widget):
        """Draws the code block"""
        del option      # unused argument
        del widget      # unused argument

        settings = self.canvas.settings
        painter.setPen(self.getPainterPen(self.isSelected(), self.borderColor))
        painter.setBrush(QBrush(self.bgColor))

        painter.drawRoundedRect(
            self.baseX + settings.hCellPadding,
            self.baseY + settings.vCellPadding,
            self.minWidth - 2 * settings.hCellPadding,
            self.minHeight - 2 * settings.vCellPadding,
            settings.returnRectRadius, settings.returnRectRadius)
        lineXBase = self.baseX + settings.hCellPadding
        lineXPos = lineXBase + self.iconItem.iconWidth() + \
                   2 * settings.hTextPadding
        lineYBase = self.baseY + settings.vCellPadding
        painter.drawLine(
            lineXPos, lineYBase,
            lineXPos, lineYBase + self.minHeight - 2 * settings.vCellPadding)

        # Draw the text in the rectangle
        pen = QPen(self.fgColor)
        painter.setFont(settings.monoFont)
        painter.setPen(pen)

        availWidth = self.minWidth - 2 * settings.hCellPadding - \
                     self.iconItem.iconWidth() - 2 * settings.hTextPadding - \
                     settings.hTextPadding - settings.returnRectRadius
        textShift = (availWidth - self.textRect.width()) / 2
        painter.drawText(
            self.baseX + settings.hCellPadding + self.iconItem.iconWidth() +
            3 * settings.hTextPadding + textShift,
            self.baseY + settings.vCellPadding + settings.vTextPadding,
            self.textRect.width(), self.textRect.height(),
            Qt.AlignLeft, self.text)

    def getLineRange(self):
        """Provides the item line range"""
        if self.ref.value is not None:
            return [self.ref.body.beginLine, self.ref.value.endLine]
        return [self.ref.body.beginLine, self.ref.body.endLine]

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.ref.value is not None:
            return [self.ref.body.begin, self.ref.value.end]
        return [self.ref.body.begin, self.ref.body.end]

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return 'Return at ' + CellElement.getLinesSuffix(self.getLineRange())


class RaiseCell(CellElement,
                TextMixin, ColorMixin, IconMixin, QGraphicsRectItem):

    """Represents a single raise statement"""

    def __init__(self, ref, canvas, x, y):
        CellElement.__init__(self, ref, canvas, x, y)
        TextMixin.__init__(self)
        ColorMixin.__init__(self, ref, self.canvas.settings.raiseBGColor,
                            self.canvas.settings.raiseFGColor,
                            self.canvas.settings.raiseBorderColor)
        IconMixin.__init__(self, canvas, 'raise.svg', 'raise')
        QGraphicsRectItem.__init__(self, canvas.scopeRectangle)
        self.kind = CellElement.RAISE
        self.connector = None

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        self.setupText(self)

        vPadding = 2 * (settings.vCellPadding + settings.vTextPadding)
        self.minHeight = self.textRect.height() + vPadding
        self.minWidth = max(
            self.textRect.width() + 2 * settings.hCellPadding +
            settings.hTextPadding + settings.returnRectRadius +
            2 * settings.hTextPadding + self.iconItem.iconWidth(),
            settings.minWidth)
        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        # Add the connector as a separate scene item to make the selection
        # working properly
        settings = self.canvas.settings
        self.connector = Connector(self.canvas, baseX + settings.mainLine,
                                   baseY,
                                   baseX + settings.mainLine,
                                   baseY + settings.vCellPadding)

        # Setting the rectangle is important for the selection and for
        # redrawing. Thus the selection pen must be considered too.
        penWidth = settings.selectPenWidth - 1
        self.setRect(baseX + settings.hCellPadding - penWidth,
                     baseY + settings.vCellPadding - penWidth,
                     self.minWidth - 2 * settings.hCellPadding + 2 * penWidth,
                     self.minHeight - 2 * settings.vCellPadding + 2 * penWidth)

        self.iconItem.setPos(
            baseX + settings.hCellPadding + settings.hTextPadding,
            baseY + self.minHeight/2 - self.iconItem.iconHeight()/2)

        scene.addItem(self.connector)
        scene.addItem(self)
        scene.addItem(self.iconItem)

    def paint(self, painter, option, widget):
        """Draws the raise statement"""
        del option      # unused argument
        del widget      # unused argument

        settings = self.canvas.settings
        painter.setPen(self.getPainterPen(self.isSelected(), self.borderColor))
        painter.setBrush(QBrush(self.bgColor))

        painter.drawRoundedRect(self.baseX + settings.hCellPadding,
                                self.baseY + settings.vCellPadding,
                                self.minWidth - 2 * settings.hCellPadding,
                                self.minHeight - 2 * settings.vCellPadding,
                                settings.returnRectRadius,
                                settings.returnRectRadius)
        lineXBase = self.baseX + settings.hCellPadding
        lineXPos = lineXBase + self.iconItem.iconWidth() + \
                   2 * settings.hTextPadding
        lineYBase = self.baseY + settings.vCellPadding
        painter.drawLine(
            lineXPos, lineYBase,
            lineXPos, lineYBase + self.minHeight - 2 * settings.vCellPadding)

        # Draw the text in the rectangle
        pen = QPen(self.fgColor)
        painter.setFont(settings.monoFont)
        painter.setPen(pen)
        availWidth = self.minWidth - 2 * settings.hCellPadding - \
                     self.iconItem.iconWidth() - 2 * settings.hTextPadding - \
                     settings.hTextPadding - settings.returnRectRadius
        textShift = (availWidth - self.textRect.width()) / 2
        painter.drawText(
            self.baseX + settings.hCellPadding + self.iconItem.iconWidth() +
            3 * settings.hTextPadding + textShift,
            self.baseY + settings.vCellPadding + settings.vTextPadding,
            self.textRect.width(), self.textRect.height(),
            Qt.AlignLeft, self.text)

    def getLineRange(self):
        """Provides the line range"""
        if self.ref.value is not None:
            return [self.ref.body.beginLine, self.ref.value.endLine]
        return [self.ref.body.beginLine, self.ref.body.endLine]

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.ref.value is not None:
            return [self.ref.body.begin, self.ref.value.end]
        return [self.ref.body.begin, self.ref.body.end]

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return 'Raise at ' + CellElement.getLinesSuffix(self.getLineRange())


class AssertCell(CellElement,
                 TextMixin, ColorMixin, IconMixin, QGraphicsRectItem):

    """Represents a single assert statement"""

    def __init__(self, ref, canvas, x, y):
        CellElement.__init__(self, ref, canvas, x, y)
        TextMixin.__init__(self)
        ColorMixin.__init__(self, ref, self.canvas.settings.assertBGColor,
                            self.canvas.settings.assertFGColor,
                            self.canvas.settings.assertBorderColor)
        IconMixin.__init__(self, canvas, 'assert.svg', 'assert')
        QGraphicsRectItem.__init__(self, canvas.scopeRectangle)
        self.kind = CellElement.ASSERT
        self.connector = None

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        self.setupText(self)

        # for an arrow box
        singleCharRect = self.getBoundingRect('W')
        self.__diamondHeight = singleCharRect.height() + \
                               2 * settings.vTextPadding
        self.__diamondWidth = settings.ifWidth * 2 + singleCharRect.width() + \
                              2 * settings.hTextPadding

        self.minHeight = self.textRect.height() + \
                         2 * settings.vCellPadding + 2 * settings.vTextPadding
        self.minWidth = max(
            self.textRect.width() + 2 * settings.hCellPadding +
            2 * settings.hTextPadding + self.__diamondWidth,
            settings.minWidth)
        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        # Add the connector as a separate scene item to make the selection
        # working properly
        settings = self.canvas.settings
        self.connector = Connector(self.canvas, baseX + settings.mainLine,
                                   baseY,
                                   baseX + settings.mainLine,
                                   baseY + self.height)
        scene.addItem(self.connector)

        # Setting the rectangle is important for the selection and for
        # redrawing. Thus the selection pen must be considered too.
        penWidth = settings.selectPenWidth - 1
        self.setRect(baseX + settings.hCellPadding - penWidth,
                     baseY + settings.vCellPadding - penWidth,
                     self.minWidth - 2 * settings.hCellPadding + 2 * penWidth,
                     self.minHeight - 2 * settings.vCellPadding + 2 * penWidth)

        settings = self.canvas.settings
        self.iconItem.setPos(
            baseX + self.__diamondWidth / 2 +
            settings.hCellPadding - self.iconItem.iconWidth() / 2,
            baseY + self.minHeight / 2 - self.iconItem.iconHeight() / 2)

        scene.addItem(self)
        scene.addItem(self.iconItem)

    def paint(self, painter, option, widget):
        """Draws the code block"""
        del option      # unused argument
        del widget      # unused argument

        settings = self.canvas.settings
        painter.setPen(self.getPainterPen(self.isSelected(), self.borderColor))
        painter.setBrush(QBrush(self.bgColor))

        dHalf = int(self.__diamondHeight / 2.0)
        dx1 = self.baseX + settings.hCellPadding
        dy1 = self.baseY + int(self.minHeight / 2)
        dx2 = dx1 + settings.ifWidth
        dy2 = dy1 - dHalf
        dx3 = dx1 + self.__diamondWidth - settings.ifWidth
        dy3 = dy2
        dx4 = dx3 + settings.ifWidth
        dy4 = dy1
        dx5 = dx3
        dy5 = dy2 + 2 * dHalf
        dx6 = dx2
        dy6 = dy5

        painter.drawPolygon(QPointF(dx1, dy1), QPointF(dx2, dy2),
                            QPointF(dx3, dy3), QPointF(dx4, dy4),
                            QPointF(dx5, dy5), QPointF(dx6, dy6))

        painter.drawRect(dx4 + 1, self.baseY + settings.vCellPadding,
                         self.minWidth - 2 * settings.hCellPadding -
                         self.__diamondWidth,
                         self.minHeight - 2 * settings.vCellPadding)


        # Draw the text in the rectangle
        pen = QPen(self.fgColor)
        painter.setFont(settings.monoFont)
        painter.setPen(pen)
        availWidth = \
            self.minWidth - 2 * settings.hCellPadding - self.__diamondWidth
        textWidth = self.textRect.width() + 2 * settings.hTextPadding
        textShift = (availWidth - textWidth) / 2
        painter.drawText(
            dx4 + settings.hTextPadding + textShift,
            self.baseY + settings.vCellPadding + settings.vTextPadding,
            self.textRect.width(), self.textRect.height(),
            Qt.AlignLeft, self.text)

    def getLineRange(self):
        """Provides the line range"""
        if self.ref.message is not None:
            return [self.ref.body.beginLine, self.ref.message.endLine]
        if self.ref.test is not None:
            return[self.ref.body.beginLine, self.ref.test.endLine]
        return [self.ref.body.beginLine, self.ref.body.endLine]

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.ref.message is not None:
            return [self.ref.body.begin, self.ref.message.end]
        if self.ref.test is not None:
            return[self.ref.body.begin, self.ref.test.end]
        return [self.ref.body.begin, self.ref.body.end]

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return 'Assert at ' + CellElement.getLinesSuffix(self.getLineRange())


class SysexitCell(CellElement,
                  TextMixin, ColorMixin, IconMixin, QGraphicsRectItem):

    """Represents a single sys.exit(...) statement"""

    def __init__(self, ref, canvas, x, y):
        CellElement.__init__(self, ref, canvas, x, y)
        TextMixin.__init__(self)
        ColorMixin.__init__(self, ref, self.canvas.settings.sysexitBGColor,
                            self.canvas.settings.sysexitFGColor,
                            self.canvas.settings.sysexitBorderColor)
        IconMixin.__init__(self, canvas, 'sysexit.svg', 'sys.exit()')
        QGraphicsRectItem.__init__(self, canvas.scopeRectangle)
        self.kind = CellElement.SYSEXIT
        self.connector = None

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        self.setupText(self)

        self.minHeight = \
            self.textRect.height() + \
            2 * (settings.vCellPadding + settings.vTextPadding)
        self.minWidth = max(
            self.textRect.width() + 2 * settings.hCellPadding +
            settings.hTextPadding + settings.returnRectRadius +
            2 * settings.hTextPadding + self.iconItem.iconWidth(),
            settings.minWidth)
        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        # Add the connector as a separate scene item to make the selection
        # working properly
        settings = self.canvas.settings
        self.connector = Connector(self.canvas, baseX + settings.mainLine,
                                   baseY,
                                   baseX + settings.mainLine,
                                   baseY + settings.vCellPadding)

        # Setting the rectangle is important for the selection and for
        # redrawing. Thus the selection pen with must be considered too.
        penWidth = settings.selectPenWidth - 1
        self.setRect(baseX + settings.hCellPadding - penWidth,
                     baseY + settings.vCellPadding - penWidth,
                     self.minWidth - 2 * settings.hCellPadding + 2 * penWidth,
                     self.minHeight - 2 * settings.vCellPadding + 2 * penWidth)

        self.iconItem.setPos(
            baseX + settings.hCellPadding + settings.hTextPadding,
            baseY + self.minHeight/2 - self.iconItem.iconHeight() / 2)

        scene.addItem(self.connector)
        scene.addItem(self)
        scene.addItem(self.iconItem)

    def paint(self, painter, option, widget):
        """Draws the sys.exit call"""
        del option      # unused argument
        del widget      # unused argument

        settings = self.canvas.settings
        painter.setPen(self.getPainterPen(self.isSelected(), self.borderColor))
        painter.setBrush(QBrush(self.bgColor))

        painter.drawRoundedRect(self.baseX + settings.hCellPadding,
                                self.baseY + settings.vCellPadding,
                                self.minWidth - 2 * settings.hCellPadding,
                                self.minHeight - 2 * settings.vCellPadding,
                                settings.returnRectRadius,
                                settings.returnRectRadius)
        lineXBase = self.baseX + settings.hCellPadding
        lineXPos = lineXBase + self.iconItem.iconWidth() + \
            2 * settings.hTextPadding
        lineYBase = self.baseY + settings.vCellPadding
        painter.drawLine(
            lineXPos, lineYBase,
            lineXPos, lineYBase + self.minHeight - 2 * settings.vCellPadding)

        # Draw the text in the rectangle
        pen = QPen(self.fgColor)
        painter.setFont(settings.monoFont)
        painter.setPen(pen)
        availWidth = \
            self.minWidth - 2 * settings.hCellPadding - \
            self.iconItem.iconWidth() - \
            2 * settings.hTextPadding - \
            settings.hTextPadding - settings.returnRectRadius
        textShift = (availWidth - self.textRect.width()) / 2
        painter.drawText(
            self.baseX + settings.hCellPadding + self.iconItem.iconWidth() +
            3 * settings.hTextPadding + textShift,
            self.baseY + settings.vCellPadding + settings.vTextPadding,
            self.textRect.width(), self.textRect.height(),
            Qt.AlignLeft, self.text)

    def getSelectTooltip(self):
        """Provides tooltip"""
        return 'Sys.exit() at ' + \
            CellElement.getLinesSuffix(self.getLineRange())


class ImportCell(CellElement,
                 TextMixin, ColorMixin, IconMixin, QGraphicsRectItem):

    """Represents a single import statement"""

    def __init__(self, ref, canvas, x, y):
        CellElement.__init__(self, ref, canvas, x, y)
        TextMixin.__init__(self)
        ColorMixin.__init__(self, ref, self.canvas.settings.importBGColor,
                            self.canvas.settings.importFGColor,
                            self.canvas.settings.importBorderColor)
        IconMixin.__init__(self, canvas, 'import.svg', 'import')
        QGraphicsRectItem.__init__(self, canvas.scopeRectangle)
        self.kind = CellElement.IMPORT
        self.connector = None

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        self.setupText(self)

        self.minHeight = \
            self.textRect.height() + 2 * settings.vCellPadding + \
            2 * settings.vTextPadding
        self.minWidth = max(
            self.textRect.width() + 2 * settings.hCellPadding +
            2 * settings.hTextPadding + self.iconItem.iconWidth() +
            2 * settings.hTextPadding, settings.minWidth)
        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        # Add the connector as a separate scene item to make the selection
        # working properly
        settings = self.canvas.settings
        self.connector = Connector(self.canvas, baseX + settings.mainLine,
                                   baseY,
                                   baseX + settings.mainLine,
                                   baseY + self.height)
        scene.addItem(self.connector)

        # Setting the rectangle is important for the selection and for
        # redrawing. Thus the selection pen must be considered too.
        penWidth = settings.selectPenWidth - 1
        self.setRect(baseX + settings.hCellPadding - penWidth,
                     baseY + settings.vCellPadding - penWidth,
                     self.minWidth - 2 * settings.hCellPadding + 2 * penWidth,
                     self.minHeight - 2 * settings.vCellPadding + 2 * penWidth)

        self.iconItem.setPos(
            baseX + settings.hCellPadding + settings.hTextPadding,
            baseY + self.minHeight / 2 - self.iconItem.iconHeight() / 2)
        scene.addItem(self)
        scene.addItem(self.iconItem)

    def paint(self, painter, option, widget):
        """Draws the import statement"""
        del option      # unused argument
        del widget      # unused argument

        settings = self.canvas.settings
        painter.setPen(self.getPainterPen(self.isSelected(), self.borderColor))
        painter.setBrush(QBrush(self.bgColor))

        painter.drawRect(self.baseX + settings.hCellPadding,
                         self.baseY + settings.vCellPadding,
                         self.minWidth - 2 * settings.hCellPadding,
                         self.minHeight - 2 * settings.vCellPadding)
        painter.drawLine(self.baseX + settings.hCellPadding +
                         self.iconItem.iconWidth() + 2 * settings.hTextPadding,
                         self.baseY + settings.vCellPadding,
                         self.baseX + settings.hCellPadding +
                         self.iconItem.iconWidth() + 2 * settings.hTextPadding,
                         self.baseY + self.minHeight - settings.vCellPadding)

        # Draw the text in the rectangle
        pen = QPen(self.fgColor)
        painter.setFont(settings.monoFont)
        painter.setPen(pen)
        textRectWidth = self.minWidth - 2 * settings.hCellPadding - \
                        4 * settings.hTextPadding - self.iconItem.iconWidth()
        textShift = (textRectWidth - self.textRect.width()) / 2
        painter.drawText(
            self.baseX + settings.hCellPadding + self.iconItem.iconWidth() +
            3 * settings.hTextPadding + textShift,
            self.baseY + settings.vCellPadding + settings.vTextPadding,
            self.textRect.width(), self.textRect.height(),
            Qt.AlignLeft, self.text)

    def getSelectTooltip(self):
        """Provides the select tooltip"""
        return 'Import at ' + CellElement.getLinesSuffix(self.getLineRange())


class IfCell(CellElement, TextMixin, ColorMixin, QGraphicsRectItem):

    """Represents a single if statement"""

    def __init__(self, ref, canvas, x, y):
        CellElement.__init__(self, ref, canvas, x, y)
        TextMixin.__init__(self)
        ColorMixin.__init__(self, ref, self.canvas.settings.ifBGColor,
                            self.canvas.settings.ifFGColor,
                            self.canvas.settings.ifBorderColor)
        QGraphicsRectItem.__init__(self, canvas.scopeRectangle)
        self.kind = CellElement.IF
        self.vConnector = None
        self.rhsConnector = None
        self.leftBadge = None
        self.yBelow = False
        self.needHConnector = True
        self.rhsShift = 0

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        self.setupText(self)

        self.minHeight = self.textRect.height() + \
                         2 * settings.vCellPadding + 2 * settings.vTextPadding
        self.minWidth = max(
            self.textRect.width() +
            2 * settings.hCellPadding + 2 * settings.hTextPadding +
            2 * settings.ifWidth,
            settings.minWidth)
        self.minWidth += self.hShift * 2 * settings.openGroupHSpacer
        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def __calcPolygon(self):
        """Calculates the polygon"""
        settings = self.canvas.settings

        shift = self.hShift * 2 * settings.openGroupHSpacer
        baseX = self.baseX + shift

        self.x1 = baseX + settings.hCellPadding
        self.y1 = self.baseY + self.minHeight / 2
        self.x2 = baseX + settings.hCellPadding + settings.ifWidth
        self.y2 = self.baseY + settings.vCellPadding
        self.x3 = baseX + self.minWidth - \
                  settings.hCellPadding - settings.ifWidth - shift
        self.y3 = self.y2
        self.x4 = self.x3 + settings.ifWidth
        self.y4 = self.y1
        self.x5 = self.x3
        self.y5 = self.baseY + (self.minHeight - settings.vCellPadding)
        self.x6 = self.x2
        self.y6 = self.y5

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        self.__calcPolygon()

        settings = self.canvas.settings
        hShift = self.hShift * 2 * settings.openGroupHSpacer
        self.baseX += hShift

        # Add the connectors as separate scene items to make the selection
        # working properly
        settings = self.canvas.settings
        self.vConnector = Connector(self.canvas,
                                    self.baseX + settings.mainLine,
                                    self.baseY,
                                    self.baseX + settings.mainLine,
                                    self.baseY + self.height)
        scene.addItem(self.vConnector)

        if self.needHConnector:
            # Need the RHS connector
            self.rhsConnector = Connector(self.canvas, self.x4, self.y4,
                                          self.baseX + self.width - hShift,
                                          self.y4)
        else:
            # Need the bottom connector because the RHS layout used some space
            # at the left
            xPos = self.baseX + self.width - self.rhsShift + settings.mainLine - hShift
            self.rhsConnector = Connector(self.canvas,
                                          xPos,
                                          self.y5,
                                          xPos,
                                          self.baseY + self.height)

        scene.addItem(self.rhsConnector)

        self.yBelow = CMLVersion.find(self.ref.leadingCMLComments,
                                      CMLsw) is not None
        if self.yBelow:
            self.leftBadge = BadgeItem(self, 'y')
            self.leftBadge.setFGColor(settings.ifYBranchTextColor)
        else:
            self.leftBadge = BadgeItem(self, 'n')
            self.leftBadge.setFGColor(settings.ifNBranchTextColor)

        self.leftBadge.setNeedRectangle(False)
        self.leftBadge.moveTo(self.x1 - self.leftBadge.width / 2,
                              self.y3 - self.leftBadge.height / 2)

        penWidth = settings.selectPenWidth - 1
        self.setRect(self.x1 - penWidth, self.y2 - penWidth,
                     self.x4 - self.x1 + 2 * penWidth,
                     self.y6 - self.y2 + 2 * penWidth)
        scene.addItem(self)
        scene.addItem(self.leftBadge)

        self.baseX -= hShift

    def paint(self, painter, option, widget):
        """Draws the code block"""
        del option      # unused argument
        del widget      # unused argument

        settings = self.canvas.settings
        painter.setPen(self.getPainterPen(self.isSelected(), self.borderColor))
        painter.setBrush(QBrush(self.bgColor))

        painter.drawPolygon(
            QPointF(self.x1, self.y1), QPointF(self.x2, self.y2),
            QPointF(self.x3, self.y3), QPointF(self.x4, self.y4),
            QPointF(self.x5, self.y5), QPointF(self.x6, self.y6))

        # Draw the text in the rectangle
        pen = QPen(self.fgColor)
        painter.setPen(pen)
        painter.setFont(settings.monoFont)
        availWidth = self.x3 - self.x2
        textWidth = self.textRect.width() + 2 * settings.hTextPadding
        textShift = (availWidth - textWidth) / 2
        painter.drawText(
            self.x2 + settings.hTextPadding + textShift,
            self.baseY + settings.vCellPadding + settings.vTextPadding,
            self.textRect.width(), self.textRect.height(),
            Qt.AlignLeft | Qt.AlignVCenter, self.text)

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return 'If at ' + CellElement.getLinesSuffix(self.getLineRange())




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""All the imports from the QT libraries"""


from PyQt5.QtCore import (Qt, QFileInfo, QSize, QUrl, QTimer, pyqtSignal,
                          QEventLoop, QRect, QEvent, QPoint, QRectF,
                          QModelIndex, QAbstractItemModel, QItemSelectionModel,
                          QStringListModel, QDir, QRegExp, QPointF, QSizeF,
                          QSortFilterProxyModel, QObject, QFileSystemWatcher,
                          QThread, QMutex, QWaitCondition, QT_VERSION_STR,
                          QMimeData, QByteArray, QProcess, QProcessEnvironment)
from PyQt5.QtGui import (QCursor, QFontMetrics, QDesktopServices, QFont, QIcon,
                         QPalette, QColor, QBrush, QKeySequence, QIntValidator,
                         QPainter, QTextCursor, QImage, QPixmap, QPen,
                         QDoubleValidator, QImageReader, QPainterPath,
                         QTransform, QTextOption, QTextDocument)
from PyQt5.QtWidgets import (QApplication, QToolBar, QMenuBar, QLabel,
                             QTabWidget, QActionGroup, QHBoxLayout, QWidget,
                             QAction, QMenu, QSizePolicy, QToolButton, QDialog,
                             QToolTip, QVBoxLayout, QSplitter, QTextBrowser,
                             QDialogButtonBox, QFrame, QGridLayout, QComboBox,
                             QAbstractItemView, QCompleter, QListView,
                             QDirModel, QMessageBox, QShortcut, QFileDialog,
                             QTabBar, QTreeView, QHeaderView, QProgressBar,
                             QCheckBox, QRadioButton, QGroupBox, QPushButton,
                             QTreeWidget, QTreeWidgetItem, QLineEdit, QStyle,
                             QListWidget, QStyleOptionFrame, QItemDelegate,
                             QStyleFactory, QStyledItemDelegate, QBoxLayout,
                             QStyleOptionViewItem, QPlainTextEdit, QTextEdit,
                             QMainWindow, QSpacerItem, QStackedWidget,
                             QSplashScreen, QFontComboBox, QScrollArea,
                             QGraphicsScene, QGraphicsSimpleTextItem,
                             QGraphicsRectItem, QGraphicsPathItem, QSpinBox,
                             QGraphicsItem, QStyleOptionGraphicsItem,
                             QGraphicsView, QGraphicsTextItem, QDockWidget,
                             QGraphicsPixmapItem, QColorDialog, QFontDialog,
                             QGraphicsLineItem)
from PyQt5.QtNetwork import QTcpServer, QHostAddress, QAbstractSocket
from PyQt5.QtSvg import QSvgGenerator, QGraphicsSvgItem
from PyQt5.QtPrintSupport import QPrinter, QPrintDialog
from PyQt5.QtPrintSupport import QPrintDialog



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""The globals viewer implementation"""

from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.project import CodimensionProject
from .qt import (Qt, QSize, QRect, QItemSelectionModel, QMenu, QWidget,
                 QAction, QVBoxLayout, QToolBar, QLabel, QSizePolicy,
                 QCursor)
from .combobox import CDMComboBox
from .globalsbrowser import GlobalsBrowser
from .viewitems import (DecoratorItemType, FunctionItemType,
                        ClassItemType, AttributeItemType,
                        GlobalItemType)


class GlobalsViewer(QWidget):

    """The globals viewer widget"""

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        self.filterEdit = None
        self.definitionButton = None
        self.findButton = None
        self.globalsViewer = None
        self.copyPathButton = None
        self.__createLayout()

        # create the context menu
        self.__menu = QMenu(self)
        self.__jumpMenuItem = self.__menu.addAction(
            getIcon('definition.png'),
            'Jump to definition', self.__goToDefinition)
        self.__menu.addSeparator()
        self.__findMenuItem = self.__menu.addAction(
            getIcon('findusage.png'), 'Find occurence', self.__findWhereUsed)
        self.__menu.addSeparator()
        self.__copyMenuItem = self.__menu.addAction(
            getIcon('copymenu.png'),
            'Copy path to clipboard', self.globalsViewer.copyToClipboard)
        self.globalsViewer.setContextMenuPolicy(Qt.CustomContextMenu)
        self.globalsViewer.customContextMenuRequested.connect(
            self.__handleShowContextMenu)

        GlobalData().project.sigProjectChanged.connect(self.__onProjectChanged)
        self.globalsViewer.sigSelectionChanged.connect(self.__selectionChanged)
        self.globalsViewer.sigOpeningItem.connect(self.itemActivated)
        self.globalsViewer.sigModelFilesChanged.connect(self.modelFilesChanged)

        self.filterEdit.lineEdit().setFocus()
        self.__contextItem = None

    def __createLayout(self):
        """Helper to create the viewer layout"""
        self.globalsViewer = GlobalsBrowser()

        # Toolbar part - buttons
        self.definitionButton = QAction(
            getIcon('definition.png'),
            'Jump to highlighted item definition', self)
        self.definitionButton.triggered.connect(self.__goToDefinition)
        self.findButton = QAction(
            getIcon('findusage.png'), 'Find highlighted item occurences', self)
        self.findButton.triggered.connect(self.__findWhereUsed)
        self.copyPathButton = QAction(
            getIcon('copymenu.png'), 'Copy path to clipboard', self)
        self.copyPathButton.triggered.connect(
            self.globalsViewer.copyToClipboard)

        self.toolbar = QToolBar(self)
        self.toolbar.setMovable(False)
        self.toolbar.setAllowedAreas(Qt.TopToolBarArea)
        self.toolbar.setIconSize(QSize(16, 16))
        self.toolbar.setContentsMargins(0, 0, 0, 0)
        self.toolbar.addAction(self.definitionButton)
        self.toolbar.addAction(self.findButton)
        self.toolbar.addAction(self.copyPathButton)

        filterLabel = QLabel("  Filter ")
        filterLabel.setStyleSheet('background: transparent')
        self.toolbar.addWidget(filterLabel)
        self.filterEdit = CDMComboBox(True, self.toolbar)
        self.filterEdit.setSizePolicy(QSizePolicy.Expanding,
                                      QSizePolicy.Expanding)
        self.filterEdit.lineEdit().setToolTip(
            "Space separated regular expressions")
        self.toolbar.addWidget(self.filterEdit)
        self.filterEdit.editTextChanged.connect(self.__filterChanged)
        self.filterEdit.itemAdded.connect(self.__filterItemAdded)
        self.filterEdit.enterClicked.connect(self.__enterInFilter)

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.toolbar)
        layout.addWidget(self.globalsViewer)

        self.setLayout(layout)

    def __filterChanged(self, text):
        """Triggers when the filter text changed"""
        self.globalsViewer.setFilter(text)
        self.globalsViewer.updateCounter()

    def __selectionChanged(self, index):
        """Handles the changed selection"""
        if index.isValid():
            self.__contextItem = self.globalsViewer.model().item(index)
        else:
            self.__contextItem = None
        self.__updateButtons()

    def getItemCount(self):
        """Provides the # of items in the model - total, not only visible"""
        return self.globalsViewer.model().sourceModel().rowCount()

    def itemActivated(self, path, line):
        """Handles the item activation"""
        del path    # unused argument
        del line    # unused argument
        self.filterEdit.addItem(self.filterEdit.lineEdit().text())

    def __filterItemAdded(self):
        """The filter item has been added"""
        project = GlobalData().project
        if project.isLoaded():
            project.findGlobalHistory = self.filterEdit.getItems()

    def __enterInFilter(self):
        """ENTER key has been clicked in the filter"""
        # check if there any records displayed
        if self.globalsViewer.model().rowCount() == 0:
            return

        # Move the focus to the list and select the first row
        self.globalsViewer.clearSelection()
        flags = QItemSelectionModel.SelectCurrent | QItemSelectionModel.Rows
        self.globalsViewer.setSelection(
            QRect(0, 0, self.globalsViewer.width(), 1), flags)
        self.globalsViewer.setFocus()

    def __onProjectChanged(self, what):
        """Triggered when a project is changed"""
        if what == CodimensionProject.CompleteProject:
            self.__contextItem = None
            self.__updateButtons()
            self.filterEdit.clear()

            project = GlobalData().project
            if project.isLoaded():
                self.filterEdit.editTextChanged.disconnect(self.__filterChanged)
                self.filterEdit.addItems(project.findGlobalHistory)
                self.filterEdit.editTextChanged.connect(self.__filterChanged)
            self.filterEdit.clearEditText()

    def __handleShowContextMenu(self, coord):
        """Show the context menu"""
        index = self.globalsViewer.indexAt(coord)
        if not index.isValid():
            return

        # This will update the __contextItem
        self.__selectionChanged(index)

        if self.__contextItem is None:
            return

        self.__jumpMenuItem.setEnabled(self.definitionButton.isEnabled())
        self.__findMenuItem.setEnabled(self.findButton.isEnabled())
        self.__copyMenuItem.setEnabled(self.copyPathButton.isEnabled())

        self.__menu.popup(QCursor.pos())

    def __goToDefinition(self):
        """Jump to definition context menu handler"""
        if self.__contextItem is not None:
            self.globalsViewer.openItem(self.__contextItem)

    def __findWhereUsed(self):
        """Find where used context menu handler"""
        if self.__contextItem is not None:
            GlobalData().mainWindow.findWhereUsed(
                self.__contextItem.getPath(),
                self.__contextItem.sourceObj)

    def __updateButtons(self):
        """Updates the toolbar buttons depending on what is selected"""
        self.definitionButton.setEnabled(False)
        self.findButton.setEnabled(False)
        self.copyPathButton.setEnabled(False)
        if self.__contextItem is None:
            return

        if self.__contextItem.itemType == DecoratorItemType:
            self.definitionButton.setEnabled(True)
            self.copyPathButton.setEnabled(True)
            return

        if self.__contextItem.itemType in [FunctionItemType, ClassItemType,
                                           AttributeItemType, GlobalItemType]:
            self.definitionButton.setEnabled(True)
            self.findButton.setEnabled(True)
            self.copyPathButton.setEnabled(True)

    def onFileUpdated(self, fileName, uuid):
        """Triggered when the file is updated"""
        del uuid    # unused argument
        self.globalsViewer.onFileUpdated(fileName)

    def modelFilesChanged(self):
        """Triggered when the source model has files added or deleted"""
        pass



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Interactive errors report using pyflakes"""

import re
from pyflakes.checker import Checker
from radon.complexity import cc_visit_ast, sorted_results
from _ast import PyCF_ONLY_AST


# Returns a dictionary:
# {lineno1: [msg1, msg2, ...],
#  lineno2: [msg3, ...]}


IGNORE_REGEXP = re.compile(r'analysis:\s*(off|disable|ignore)', re.IGNORECASE)


def getBufferErrors(sourceCode):
    """Provides a list of warnings/errors for the given source code"""
    sourceCode += '\n'

    # First, compile into an AST and handle syntax errors.
    try:
        tree = compile(sourceCode, "<string>", "exec", PyCF_ONLY_AST)
    except SyntaxError as value:
        # If there's an encoding problem with the file, the text is None.
        if value.text is None:
            return {}, []
        return {value.lineno: [value.args[0]]}, []
    except (ValueError, TypeError) as value:
        # ValueError may happened in case of invalid \x escape character
        # E.g. http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=674797
        # TypeError may happened in case of null characters in a file
        # E.g. http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=674796
        msg = str(value)
        if msg == "":
            return {-1: ["Could not compile buffer: unknown error"]}, []
        return {-1: ["Could not compile buffer: " + msg]}, []

    # Okay, it's syntactically valid.  Now check it.
    check = Checker(tree, "<string>")
    results = {}
    lines = sourceCode.splitlines()
    for warning in check.messages:
        if isinstance(warning.lineno, int):
            lineno = warning.lineno
        else:
            # By some reasons I see ast NAME node here (pyflakes 0.7.3)
            lineno = warning.lineno.lineno
        if not IGNORE_REGEXP.search(lines[lineno - 1]):
            if lineno in results:
                results[lineno].append(warning.message % warning.message_args)
            else:
                results[lineno] = [warning.message % warning.message_args]

    # Radon: CC complexity as the second value
    return results, sorted_results(cc_visit_ast(tree))



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Collection of items which may appear in various tree views:
   - project view
   - filesystem view
   - classes view
   - functions view
"""

import os
import os.path
from utils.pixmapcache import getIcon
from utils.fileutils import getFileProperties, isPythonMime, isCDMProjectMime
from utils.project import getProjectFileTooltip
from .qt import Qt


NoItemType = -1

DirectoryItemType = 1
SysPathItemType = 2
FileItemType = 3
GlobalsItemType = 4
ImportsItemType = 5
FunctionsItemType = 6
ClassesItemType = 7
StaticAttributesItemType = 8
InstanceAttributesItemType = 9

CodingItemType = 20
ImportItemType = 21
ImportWhatItemType = 22
FunctionItemType = 23
ClassItemType = 24
DecoratorItemType = 25
AttributeItemType = 26
GlobalItemType = 27

EMPTY_ICON = getIcon('empty.png')


class TreeViewItem():

    """Common data structures for tree views items"""

    def __init__(self, parent, data):

        self.childItems = []
        self.childItemsSize = 0

        if isinstance(data, list):
            self.itemData = data
            self.itemDataSize = len(self.itemData)
        else:
            self.itemData = [data]
            self.itemDataSize = 1

        self.isLink = False
        self.parentItem = parent
        self.itemType = NoItemType
        self.icon = EMPTY_ICON
        self.populated = True
        self.lazyPopulation = False
        self.toolTip = ""
        self.path = None
        self.needVCSStatus = False

        self.vcsStatus = None   # VCSStatus instance, if present

    def isRoot(self):
        """True if it is the root item"""
        return self.parentItem is None

    def appendData(self, data):
        """Adds more data to the item"""
        if isinstance(data, list):
            self.itemData += data
            self.itemDataSize = len(self.itemData)
        else:
            self.itemData.append(data)
            self.itemDataSize += 1

    def setPath(self, path):
        """Sets the item path"""
        self.path = path

    def appendChild(self, child):
        """Adds a child to the item"""
        self.childItems.append(child)
        self.childItemsSize += 1
        self.populated = True

    def removeChild(self, child):
        """Removes the child"""
        self.childItems.remove(child)
        self.childItemsSize -= 1

    def removeChildren(self):
        """Removes all children"""
        self.childItems = []
        self.childItemsSize = 0

    def child(self, row):
        """Provides the child"""
        return self.childItems[row]

    def children(self):
        """Provides all the child items"""
        return self.childItems[:]

    def childCount(self):
        """Provides the number of child items"""
        # Optimization against len(self.childItems)
        return self.childItemsSize

    def columnCount(self):
        """Provides the number of available data items"""
        # Optimization against len(self.itemData)
        return self.itemDataSize

    def data(self, column):
        """Provides item data"""
        try:
            return self.itemData[column]
        except IndexError:
            return ''

    def setData(self, column, value):
        """Sets the new data"""
        try:
            self.itemData[column] = value
        except:
            pass

    def parent(self):
        """Provides the reference to the parent item"""
        return self.parentItem

    def row(self):
        """Provides the row number of this item"""
        return self.parentItem.childItems.index(self)

    def type(self):
        """Provides the item type"""
        return self.itemType

    def getIcon(self):
        """Provides the items icon"""
        return self.icon

    def setIcon(self, icon):
        """Sets the icon"""
        self.icon = icon

    def lessThan(self, other, column, order):
        """Check, if the item is less than another"""
        try:
            return self.itemData[column] < other.itemData[column]
        except IndexError:
            return False

    def getPath(self):
        """Provides the file name or dir name depending on the item type"""
        if self.itemType in [SysPathItemType, NoItemType]:
            return ''

        current = self

        # The item could be used for G/F/C viewers where the file name is
        # available in the second column
        if self.path:
            return self.path
        if self.itemDataSize >= 2:
            return self.itemData[1]

        # Memorize if it is a directory item
        isDir = current.itemType == DirectoryItemType

        # skip till file or directory items
        while True:
            if current.itemType == NoItemType:
                raise Exception("Internal error of getting tree view "
                                "item path. Please inform the developers.")
            if current.itemType in [FileItemType, DirectoryItemType,
                                    SysPathItemType]:
                break
            current = current.parentItem

        path = current.itemData[0]
        while current.parentItem.itemType != NoItemType and \
              current.parentItem.itemType != SysPathItemType:
            current = current.parentItem
            path = os.path.sep + path
            if current.itemData[0] != os.path.sep:    # root item
                path = current.itemData[0] + path

        path = os.path.normpath(path)
        if isDir:
            if not path.endswith(os.path.sep):
                path += os.path.sep
        return path

    def getRealPath(self):

        """Provides the file name or dir name depending on the item type.
           All the links except of the item itself are resolved.
        """
        if self.itemType in [SysPathItemType, NoItemType]:
            return ''

        current = self

        # The item could be used for G/F/C viewers where the file name is
        # available in the second column
        if self.path:
            return self.path
        if self.itemDataSize >= 2:
            return self.itemData[1]

        # Memorize if it is a directory item
        isDir = self.itemType == DirectoryItemType

        # skip till file or directory items
        while True:
            if current.itemType == NoItemType:
                raise Exception("Internal error of getting tree view "
                                "item path. Please inform the developers.")
            if current.itemType in [FileItemType, DirectoryItemType,
                                    SysPathItemType]:
                break
            current = current.parentItem

        hasLinks = False
        path = current.itemData[0]
        while current.parentItem.itemType != NoItemType and \
              current.parentItem.itemType != SysPathItemType:
            current = current.parentItem
            if current.itemType == DirectoryItemType:
                if current.isLink:
                    hasLinks = True
            path = os.path.sep + path
            if current.itemData[0] != os.path.sep:    # root item
                path = current.itemData[0] + path

        if hasLinks:
            if self.isLink:
                path = os.path.realpath(os.path.dirname(path)) + \
                       os.path.sep + self.itemData[0]
            else:
                path = os.path.realpath(path)

        path = os.path.normpath(path)
        if isDir:
            if not path.endswith(os.path.sep):
                path += os.path.sep
        return path

    def getQualifiedName(self):
        """Provides the qualified item name"""
        current = self
        name = ""
        while current is not None:
            if current.itemType in [FunctionItemType, ClassItemType]:
                namePart = current.data(0).split("(")[0]
                if name != "":
                    name = namePart + "." + name
                else:
                    name = namePart
            current = current.parentItem
        return name

    def getRowPath(self):
        """Provides the row path"""
        rowPath = []
        child = self
        current = self.parentItem
        while not current is None:
            for index in range(0, len(current.childItems)):
                if current.childItems[index] == child:
                    rowPath = [index] + rowPath
                    child = current
                    current = child.parentItem
                    break
        return rowPath

    def getDisplayDataPath(self):
        """Provides the diplayed path for the item"""
        result = []
        current = self
        while current.parentItem is not None:
            result.insert(0, (current.itemType, current.data(0)))
            current = current.parentItem
        return result


class TreeViewDirectoryItem(TreeViewItem):

    """Directory item"""

    def __init__(self, parent, dinfo, full=True):

        self._dirName = os.path.abspath(str(dinfo))

        if full:
            dname = self._dirName
        else:
            dname = os.path.basename(self._dirName)

        TreeViewItem.__init__(self, parent, dname)

        self.itemType = DirectoryItemType

        self.icon = None
        self.populated = False
        self.lazyPopulation = True
        self.isLink = False
        self.updateStatus()

    def updateStatus(self):
        """Updates internal fields"""
        if os.path.exists(self._dirName):
            self.icon = getIcon('dirclosed.png')
            self.populated = False
            self.lazyPopulation = True

            if os.path.islink(self._dirName):
                self.isLink = True
                linkTo = os.readlink(self._dirName)
                realpath = os.path.realpath(self._dirName)
                self.toolTip = "-> " + linkTo + "  (" + realpath + ")"
                self.icon = getIcon('dirlink.png')
        else:
            self.icon = getIcon('dirbroken.png')
            self.populated = True
            self.lazyPopulation = False

            self.childItems = []
            self.childItemsSize = 0

    def lessThan(self, other, column, order):
        """Checks if the item is less than another"""
        if other.itemType == FileItemType:
            return order == Qt.AscendingOrder
        return TreeViewItem.lessThan(self, other, column, order)


class TreeViewSysPathItem(TreeViewItem):

    """sys.path files item"""

    def __init__(self, parent):

        TreeViewItem.__init__(self, parent, "sys.path")

        self.itemType = SysPathItemType
        self.icon = getIcon('filepython.png')
        self.populated = False
        self.lazyPopulation = True
        self.isLink = False


class TreeViewFileItem(TreeViewItem):

    """file item"""

    def __init__(self, parent, path):
        path = str(path)
        TreeViewItem.__init__(self, parent, os.path.basename(path))
        self.itemType = FileItemType
        self.parsingErrors = False  # Used for python files only
        self.isLink = False

        self.fileType, self.icon, _ = getFileProperties(path)
        if self.fileType is None:
            if self.icon is None:
                self.icon = getIcon('filemisc.png')
            return

        if 'broken-symlink' in self.fileType:
            self.isLink = True
            self.toolTip = self.__brokenLinkTooltip(path)
            return

        if os.path.islink(path):
            self.isLink = True
            self.toolTip = self.__linkTooltip(path)
            self.icon = getIcon('filelink.png')
            self.fileType, _, _ = getFileProperties(os.path.realpath(path))
            return

        # Fine corrections for some file types
        if isPythonMime(self.fileType):
            self.populated = False
            self.lazyPopulation = True
            return

        if isCDMProjectMime(self.fileType):
            # Get the project properties
            try:
                self.toolTip = getProjectFileTooltip(path)
            except:
                # cannot get project properties
                self.toolTip = 'Broken project file'
            return

    def lessThan(self, other, column, order):
        """Checks if the item is less than another"""
        if other.itemType != FileItemType:
            return order == Qt.DescendingOrder

        sinit = self.data(0).startswith('__init__.py')
        oinit = other.data(0).startswith('__init__.py')
        if sinit and not oinit:
            return order == Qt.AscendingOrder
        if not sinit and oinit:
            return order == Qt.DescendingOrder
        return TreeViewItem.lessThan(self, other, column, order)

    def updateLinkStatus(self, path):
        """Called to update the status to/from broken link"""
        if not self.isLink:
            return

        self.fileType, self.icon, _ = getFileProperties(path)
        if 'broken-symlink' in self.fileType:
            self.toolTip = self.__brokenLinkTooltip(path)
            return

        self.toolTip = self.__linkTooltip(path)
        self.icon = getIcon('filelink.png')
        self.fileType, _, _ = getFileProperties(os.path.realpath(path))

    @staticmethod
    def __brokenLinkTooltip(path):
        """Provides the broken link tooltip"""
        linkTo = os.readlink(path)
        realpath = os.path.realpath(path)
        return "Broken symlink -> " + linkTo + " (" + realpath + ")"

    @staticmethod
    def __linkTooltip(path):
        """Provides the link tooltip"""
        linkTo = os.readlink(path)
        realpath = os.path.realpath(path)
        return "-> " + linkTo + "  (" + realpath + ")"


class TreeViewGlobalsItem(TreeViewItem):

    """Globals item"""

    def __init__(self, parent, infoObj):
        TreeViewItem.__init__(self, parent, "Globals")

        self.sourceObj = infoObj
        self.itemType = GlobalsItemType
        self.icon = getIcon('globalvar.png')
        self.populated = False
        self.lazyPopulation = True

    def updateData(self, infoObj):
        """Updates data model source"""
        self.sourceObj = infoObj


class TreeViewImportsItem(TreeViewItem):

    """Imports item"""

    def __init__(self, parent, infoObj):

        TreeViewItem.__init__(self, parent, "Imports")

        self.sourceObj = infoObj
        self.itemType = ImportsItemType
        self.icon = getIcon('imports.png')
        self.populated = False
        self.lazyPopulation = True

    def updateData(self, infoObj):
        """Updates data model source"""
        self.sourceObj = infoObj


class TreeViewFunctionsItem(TreeViewItem):

    """Functions item"""

    def __init__(self, parent, infoObj):
        TreeViewItem.__init__(self, parent, "Functions")

        self.sourceObj = infoObj
        self.itemType = FunctionsItemType
        self.icon = getIcon('method.png')
        self.populated = False
        self.lazyPopulation = True

    def updateData(self, infoObj):
        """Updates data model source"""
        self.sourceObj = infoObj


class TreeViewClassesItem(TreeViewItem):

    """Classes item"""

    def __init__(self, parent, infoObj):
        TreeViewItem.__init__(self, parent, "Classes")

        self.sourceObj = infoObj
        self.itemType = ClassesItemType
        self.icon = getIcon('class.png')
        self.populated = False
        self.lazyPopulation = True

    def updateData(self, infoObj):
        """Updates data model source"""
        self.sourceObj = infoObj


class TreeViewStaticAttributesItem(TreeViewItem):

    """Static attributes item"""

    def __init__(self, parent):
        TreeViewItem.__init__(self, parent, "Static attributes")

        self.itemType = StaticAttributesItemType
        self.icon = getIcon('attributes.png')
        self.populated = False
        self.lazyPopulation = True


class TreeViewInstanceAttributesItem(TreeViewItem):

    """Instance attributes item"""

    def __init__(self, parent):

        TreeViewItem.__init__(self, parent, "Instance attributes")

        self.itemType = InstanceAttributesItemType
        self.icon = getIcon('attributes.png')
        self.populated = False
        self.lazyPopulation = True


class TreeViewCodingItem(TreeViewItem):

    """coding item"""

    def __init__(self, parent, encodingObj):
        TreeViewItem.__init__(self, parent, encodingObj.name)

        self.sourceObj = encodingObj
        self.itemType = CodingItemType
        self.icon = getIcon('textencoding.png')

    def updateData(self, encodingObj):
        """Updates data model source"""
        self.sourceObj = encodingObj
        self.setData(0, encodingObj.name)


class TreeViewImportItem(TreeViewItem):

    """Single import item"""

    def __init__(self, parent, importObj):
        TreeViewItem.__init__(self, parent, importObj.getDisplayName())

        self.sourceObj = importObj
        self.itemType = ImportItemType
        self.icon = getIcon('imports.png')

        self.populated = False
        self.lazyPopulation = True

    def updateData(self, importObj):
        """Updates data model source"""
        self.sourceObj = importObj
        self.setData(0, importObj.getDisplayName())


class TreeViewWhatItem(TreeViewItem):

    """Single what imported item"""

    def __init__(self, parent, whatObj):
        TreeViewItem.__init__(self, parent, whatObj.getDisplayName())

        self.sourceObj = whatObj
        self.itemType = ImportWhatItemType
        self.icon = getIcon('importwhat.png')

    def updateData(self, whatObj):
        """Updates data model source"""
        self.sourceObj = whatObj
        self.setData(0, whatObj.getDisplayName())


class TreeViewFunctionItem(TreeViewItem):

    """Single function / class  method item"""

    def __init__(self, parent, functionObj):
        TreeViewItem.__init__(self, parent, functionObj.getDisplayName())

        self.sourceObj = functionObj
        self.itemType = FunctionItemType

        self.__updateTooltip()

        # It looks better without an icon at all
        self.icon = None
        # if functionObj.isPrivate():
        #     self.icon = getIcon('method_private.png')
        # elif functionObj.isProtected():
        #     self.icon = getIcon('method_protected.png')
        # else:
        #     self.icon = getIcon('method.png')

        self.populated = False
        self.lazyPopulation = True

    def updateData(self, functionObj):
        """Updates data model source"""
        self.sourceObj = functionObj
        self.setData(0, functionObj.getDisplayName())
        self.__updateTooltip()

    def __updateTooltip(self):
        """Sets the tooltip value"""
        self.toolTip = ""
        if self.sourceObj.docstring is not None:
            self.toolTip = self.sourceObj.docstring.text


class TreeViewClassItem(TreeViewItem):

    """Single class item"""

    def __init__(self, parent, classObj):
        TreeViewItem.__init__(self, parent, classObj.getDisplayName())

        self.sourceObj = classObj
        self.itemType = ClassItemType

        self.__updateTooltip()

        # It looks better without an icon at all
        self.icon = None
        # if classObj.isPrivate():
        #     self.icon = getIcon('class_private.png')
        # elif classObj.isProtected():
        #     self.icon = getIcon('class_protected.png')
        # else:
        #     self.icon = getIcon('class.png')

        # Decide if it should be expandable
        if classObj.decorators or \
           classObj.functions or \
           classObj.classes or \
           classObj.classAttributes or \
           classObj.instanceAttributes:
            self.populated = False
            self.lazyPopulation = True

    def updateData(self, classObj):
        """Updates data model source"""
        self.sourceObj = classObj
        self.setData(0, classObj.getDisplayName())
        self.__updateTooltip()

    def __updateTooltip(self):
        """Sets the tooltip value"""
        self.toolTip = ""
        if self.sourceObj.docstring is not None:
            self.toolTip = self.sourceObj.docstring.text


class TreeViewDecoratorItem(TreeViewItem):

    """Single decorator item"""

    def __init__(self, parent, decoratorObj):
        TreeViewItem.__init__(self, parent, decoratorObj.getDisplayName())

        self.sourceObj = decoratorObj
        self.itemType = DecoratorItemType
        self.icon = getIcon('decorator.png')

    def updateData(self, decoratorObj):
        """Updates data model source"""
        self.sourceObj = decoratorObj
        self.setData(0, decoratorObj.getDisplayName())


class TreeViewAttributeItem(TreeViewItem):

    """Single attribute item"""

    def __init__(self, parent, attributeObj):
        TreeViewItem.__init__(self, parent, attributeObj.name)

        self.sourceObj = attributeObj
        self.itemType = AttributeItemType
        self.__setIcon()

    def updateData(self, attributeObj):
        """Updates data model source"""
        self.sourceObj = attributeObj
        self.setData(0, attributeObj.name)
        self.__setIcon()

    def __setIcon(self):
        """Sets the icon depending on access type"""
        if self.sourceObj.isPrivate():
            self.icon = getIcon('attribute_private.png')
        elif self.sourceObj.isProtected():
            self.icon = getIcon('attribute_protected.png')
        else:
            self.icon = getIcon('attribute.png')


class TreeViewGlobalItem(TreeViewItem):

    """Single global var item"""

    def __init__(self, parent, globalObj):
        TreeViewItem.__init__(self, parent, globalObj.name)

        self.sourceObj = globalObj
        self.itemType = GlobalItemType
        self.__setIcon()

    def updateData(self, globalObj):
        """Updates data model source"""
        self.sourceObj = globalObj
        self.setData(0, globalObj.name)
        self.__setIcon()

    def __setIcon(self):
        """Sets the icon depending on access type"""
        # It looks better without an icon at all
        self.icon = None

        # if self.sourceObj.isPrivate():
        #     self.icon = getIcon('attribute_private.png')
        # elif self.sourceObj.isProtected():
        #     self.icon = getIcon('attribute_protected.png')
        # else:
        #     self.icon = getIcon('attribute.png')




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Stack viewer"""

import os.path
from ui.qt import (Qt, QSizePolicy, QFrame, QTreeWidget, QToolButton,
                   QTreeWidgetItem, QHeaderView, QVBoxLayout, QWidget,
                   QAbstractItemView, QMenu, QToolBar, QCursor, QSize)
from ui.itemdelegates import NoOutlineHeightDelegate
from ui.labels import HeaderFitLabel
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.settings import Settings


class StackFrameItem(QTreeWidgetItem):

    """Single stack frame item data structure"""

    def __init__(self, fileName, lineNumber, funcName, funcArgs, frameNumber):
        shortened = os.path.basename(fileName) + ":" + str(lineNumber)

        self.__lineNumber = lineNumber
        QTreeWidgetItem.__init__(self, ["", shortened,
                                        funcName, funcArgs, fileName])

        self.__isCurrent = False
        self.__frameNumber = frameNumber

        tooltip = ['Location: ' + fileName + ':' + str(lineNumber)]
        if funcName:
            tooltip += ['Function: ' + funcName,
                        'Arguments: ' + funcArgs]
        self.setToolTip(0, '\n'.join(tooltip))

    def setCurrent(self, value):
        """Mark the current stack frame with an icon if so"""
        self.__isCurrent = value
        if value:
            self.setIcon(0, getIcon('currentframe.png'))
        else:
            self.setIcon(0, getIcon('empty.png'))

    def getFrameNumber(self):
        """Provides the frame number"""
        return self.__frameNumber

    def getFilename(self):
        """Provides the full project filename"""
        return str(self.text(4))

    def getLineNumber(self):
        """Provides the line number"""
        return self.__lineNumber

    def isCurrent(self):
        """True if the project is current"""
        return self.__isCurrent


class StackViewer(QWidget):

    """Implements the stack viewer for a debugger"""

    def __init__(self, debugger, parent=None):
        QWidget.__init__(self, parent)

        self.__debugger = debugger
        self.currentStack = None
        self.currentFrame = 0
        self.__contextItem = None
        self.__createPopupMenu()
        self.__createLayout()

    def __createPopupMenu(self):
        """Creates the popup menu"""
        self.__framesMenu = QMenu()
        self.__setCurrentMenuItem = self.__framesMenu.addAction(
            "Set current (single click)", self.__onSetCurrent)
        self.__jumpMenuItem = self.__framesMenu.addAction(
            "Set current and jump to the source (double click)",
            self.__onSetCurrentAndJump)

    def __createLayout(self):
        """Creates the widget layout"""
        verticalLayout = QVBoxLayout(self)
        verticalLayout.setContentsMargins(0, 0, 0, 0)
        verticalLayout.setSpacing(0)

        self.__stackLabel = HeaderFitLabel(self)
        self.__stackLabel.setText('Stack')
        self.__stackLabel.setSizePolicy(QSizePolicy.Expanding,
                                        QSizePolicy.Fixed)
        self.__stackLabel.setMinimumWidth(10)

        self.__showHideButton = QToolButton()
        self.__showHideButton.setAutoRaise(True)
        self.__showHideButton.setIcon(getIcon('less.png'))
        self.__showHideButton.setFixedSize(self.__stackLabel.height(),
                                           self.__stackLabel.height())
        self.__showHideButton.setToolTip('Hide frames list')
        self.__showHideButton.setFocusPolicy(Qt.NoFocus)
        self.__showHideButton.clicked.connect(self.__onShowHide)

        self.headerToolbar = QToolBar(self)
        self.headerToolbar.setIconSize(QSize(16, 16))
        self.headerToolbar.setContentsMargins(1, 1, 1, 1)
        self.headerToolbar.addWidget(self.__stackLabel)
        self.headerToolbar.addWidget(self.__showHideButton)

        self.__framesList = QTreeWidget(self)
        self.__framesList.setSortingEnabled(False)
        # I might not need that because of two reasons:
        # - the window has no focus
        # - the window has custom current indicator
        # self.__framesList.setAlternatingRowColors(True)
        self.__framesList.setRootIsDecorated(False)
        self.__framesList.setItemsExpandable(False)
        self.__framesList.setUniformRowHeights(True)
        self.__framesList.setSelectionMode(QAbstractItemView.NoSelection)
        self.__framesList.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.__framesList.setItemDelegate(NoOutlineHeightDelegate(4))
        self.__framesList.setFocusPolicy(Qt.NoFocus)
        self.__framesList.setContextMenuPolicy(Qt.CustomContextMenu)

        self.__framesList.itemClicked.connect(self.__onFrameClicked)
        self.__framesList.itemDoubleClicked.connect(
            self.__onFrameDoubleClicked)
        self.__framesList.customContextMenuRequested.connect(
            self.__showContextMenu)

        self.__framesList.setHeaderLabels(['', 'File:line',
                                           'Function', 'Arguments',
                                           'Full path'])

        verticalLayout.addWidget(self.headerToolbar)
        verticalLayout.addWidget(self.__framesList)

    def __onShowHide(self, startup=False):
        """Triggered when show/hide button is clicked"""
        if startup or self.__framesList.isVisible():
            self.__minH = self.minimumHeight()
            self.__maxH = self.maximumHeight()
            self.splitterSize = self.parent().sizes()[1]

            self.__framesList.setVisible(False)
            self.__showHideButton.setIcon(getIcon('more.png'))
            self.__showHideButton.setToolTip('Show frames list')

            self.setMinimumHeight(self.headerToolbar.height())
            self.setMaximumHeight(self.headerToolbar.height())
        else:
            self.__framesList.setVisible(True)
            self.__showHideButton.setIcon(getIcon('less.png'))
            self.__showHideButton.setToolTip('Hide frames list')

            self.setMinimumHeight(self.__minH)
            self.setMaximumHeight(self.__maxH)
            splitterSizes = self.parent().sizes()
            splitterSizes[1] = self.splitterSize
            self.parent().setSizes(splitterSizes)

    def clear(self):
        """Clears the content"""
        self.__framesList.clear()
        self.currentStack = None
        self.__stackLabel.setText("Stack")

    def __resizeColumns(self):
        """Resize the files list columns"""
        self.__framesList.header().setStretchLastSection(True)
        self.__framesList.header().resizeSections(
            QHeaderView.ResizeToContents)
        self.__framesList.header().resizeSection(0, 22)
        self.__framesList.header().setSectionResizeMode(0, QHeaderView.Fixed)

    def populate(self, stack):
        """Sets the new call stack and selects the first item in it"""
        self.clear()

        self.currentStack = stack
        self.currentFrame = 0
        frameNumber = 0
        for item in stack:
            fName = item[0]
            lineNo = item[1]
            funcName = ''
            funcArgs = ''
            if len(item) >= 3:
                funcName = item[2]
            if len(item) >= 4:
                funcArgs = item[3]

            if funcName.startswith('<'):
                funcName = ''
                funcArgs = ''

            item = StackFrameItem(fName, lineNo,
                                  funcName, funcArgs, frameNumber)
            self.__framesList.addTopLevelItem(item)
            frameNumber += 1
        self.__resizeColumns()
        self.__framesList.topLevelItem(0).setCurrent(True)
        self.__stackLabel.setText("Stack (total: " +
                                  str(len(stack)) + ")")

    def getFrameNumber(self):
        """Provides the current frame number"""
        return self.currentFrame

    def __onFrameClicked(self, item, column):
        """Triggered when a frame is clicked"""
        del column  # unused argument
        if item.isCurrent():
            return

        # Hide the current indicator
        self.__framesList.topLevelItem(self.currentFrame).setCurrent(False)

        # Show the new indicator
        self.currentFrame = item.getFrameNumber()
        for index in range(self.__framesList.topLevelItemCount()):
            item = self.__framesList.topLevelItem(index)
            if item.getFrameNumber() == self.currentFrame:
                item.setCurrent(True)
        self.__debugger.remoteClientVariables(1, self.currentFrame)  # globals
        self.__debugger.remoteClientVariables(0, self.currentFrame)  # locals

    def __onFrameDoubleClicked(self, item, column):
        """Triggered when a frame is double clicked"""
        del column  # unused argument
        # The frame has been switched already because the double click
        # signal always comes after the single click one
        fileName = item.getFilename()
        lineNumber = item.getLineNumber()

        editorsManager = GlobalData().mainWindow.editorsManager()
        editorsManager.openFile(fileName, lineNumber)
        editor = editorsManager.currentWidget().getEditor()
        editor.gotoLine(lineNumber)
        editorsManager.currentWidget().setFocus()

    def __showContextMenu(self, coord):
        """Shows the frames list context menu"""
        self.__contextItem = self.__framesList.itemAt(coord)
        if self.__contextItem is not None:
            self.__setCurrentMenuItem.setEnabled(
                not self.__contextItem.isCurrent())
            self.__framesMenu.popup(QCursor.pos())

    def __onSetCurrent(self):
        """Context menu item handler"""
        self.__onFrameClicked(self.__contextItem, 0)

    def __onSetCurrentAndJump(self):
        """Context menu item handler"""
        self.__onFrameClicked(self.__contextItem, 0)
        self.__onFrameDoubleClicked(self.__contextItem, 0)

    def switchControl(self, isInIDE):
        """Switches the UI depending where the control flow is"""
        self.__framesList.setEnabled(isInIDE)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Profiling results as a graph"""

from subprocess import PIPE, Popen
import os
import os.path
import math
import io
import gprof2dot
from ui.qt import (Qt, QPointF, pyqtSignal, QFont, QPen, QColor,
                   QFontMetrics, QPainterPath, QPainter, QWidget, QLabel,
                   QVBoxLayout, QGraphicsScene, QGraphicsPathItem,
                   QStyle, QGraphicsTextItem, QStyleOptionGraphicsItem,
                   QGraphicsItem, QGraphicsRectItem, QSizePolicy)
from ui.labels import HeaderFitLabel
from utils.settings import Settings
from utils.globals import GlobalData
from utils.pixmapcache import getPixmap
from diagram.plaindotparser import getGraphFromPlainDotData
from diagram.importsdgmgraphics import DiagramWidget
from .proftable import FLOAT_FORMAT


DEFAULT_FONT = QFont("Arial", 10)

class FuncConnectionLabel(QGraphicsTextItem):

    """Connector label"""

    def __init__(self, edge, edgeFont):
        text = edge.label.replace('\\n', '\n')
        QGraphicsTextItem.__init__(self, text)

        if edgeFont:
            self.setFont(edgeFont)
        else:
            self.setFont(DEFAULT_FONT)

        metric = QFontMetrics(self.font())
        rec = metric.boundingRect(0, 0, 10000, 10000, Qt.AlignCenter, text)

        self.setPos(edge.labelX - rec.width() / 2,
                    edge.labelY - rec.height() / 2)

        # To make double click not delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, False)

    def paint(self, painter, option, widget):
        """ Draws the edge text """
        # Hide the dotted outline
        itemOption = QStyleOptionGraphicsItem(option)
        if itemOption.state & QStyle.State_Selected != 0:
            itemOption.state = itemOption.state & ~QStyle.State_Selected

        QGraphicsTextItem.paint(self, painter, itemOption, widget)


class FuncConnection(QGraphicsPathItem):

    """Connection between functions"""

    def __init__(self, edge):
        QGraphicsPathItem.__init__(self)
        self.__edge = edge

        startPoint = QPointF(edge.points[0][0], edge.points[0][1])
        painterPath = QPainterPath(startPoint)

        index = 1
        while index + 3 <= len(edge.points):
            painterPath.cubicTo(
                edge.points[index][0], edge.points[index][1],
                edge.points[index + 1][0], edge.points[index + 1][1],
                edge.points[index + 2][0], edge.points[index + 2][1])
            index = index + 3
        if index + 2 <= len(edge.points):
            painterPath.quadTo(
                edge.points[index + 1][0], edge.points[index + 1][1],
                edge.points[index + 2][0], edge.points[index + 2][1])
            index = index + 2

        if index + 1 <= len(edge.points):
            painterPath.lineTo(edge.points[index + 1][0],
                               edge.points[index + 1][1])

        if edge.head != edge.tail:
            lastIndex = len(edge.points) - 1
            self.addArrow(painterPath,
                          edge.points[lastIndex - 1][0],
                          edge.points[lastIndex - 1][1],
                          edge.points[lastIndex][0],
                          edge.points[lastIndex][1])
        self.setPath(painterPath)

    def addArrow(self, painterPath, startX, startY, endX, endY):
        """Add arrows to the edges
           http://kapo-cpp.blogspot.com/2008/10/drawing-arrows-with-cairo.html
        """
        arrowLength = 12.0
        arrowDegrees = 0.15      # Radian

        angle = math.atan2(endY - startY, endX - startX) + math.pi
        arrowX1 = endX + arrowLength * math.cos(angle - arrowDegrees)
        arrowY1 = endY + arrowLength * math.sin(angle - arrowDegrees)
        arrowX2 = endX + arrowLength * math.cos(angle + arrowDegrees)
        arrowY2 = endY + arrowLength * math.sin(angle + arrowDegrees)

        painterPath.moveTo(endX, endY)
        painterPath.lineTo(arrowX1, arrowY1)
        painterPath.moveTo(endX, endY)
        painterPath.lineTo(arrowX2, arrowY2)

    def paint(self, painter, option, widget):
        """Draws a curve and then adds an arrow"""
        pen = QPen(QColor(0, 0, 0))
        pen.setWidth(2)
        self.setPen(pen)

        painter.setRenderHint(QPainter.Antialiasing, True)
        QGraphicsPathItem.paint(self, painter, option, widget)


class Function(QGraphicsRectItem):

    """Rectangle for a function"""

    def __init__(self, node, fileName, lineNumber, outside, nodeFont):
        QGraphicsRectItem.__init__(self)
        self.__node = node
        self.__fileName = fileName
        self.__lineNumber = lineNumber
        self.__outside = outside

        self.__font = DEFAULT_FONT
        if nodeFont:
            self.__font = nodeFont

        posX = node.posX - node.width / 2.0
        posY = node.posY - node.height / 2.0
        QGraphicsRectItem.__init__(self, posX, posY,
                                   node.width, node.height)

        self.setRectanglePen()

        # node.color is like "#fe0400"
        if node.color.startswith("#"):
            color = QColor(int(node.color[1:3], 16),
                           int(node.color[3:5], 16),
                           int(node.color[5:], 16))
        else:
            color = QColor(220, 255, 220)
        self.setBrush(color)

        # To make item selectable
        self.setFlag(QGraphicsItem.ItemIsSelectable,
                     os.path.isabs(self.__fileName) and
                     self.__lineNumber > 0)

        # Set tooltip as a function docstring
        if fileName != "" and lineNumber != 0:
            self.setToolTip(
                GlobalData().getFileLineDocstring(fileName, lineNumber))

    def setRectanglePen(self):
        """Sets the diagram pen"""
        pen = QPen(QColor(0, 0, 0))
        pen.setWidth(2)
        pen.setJoinStyle(Qt.RoundJoin)
        self.setPen(pen)

    def paint(self, painter, option, widget):
        """Draws a filled rectangle and then adds a title"""
        # Hide the dotted outline
        itemOption = QStyleOptionGraphicsItem(option)
        if itemOption.state & QStyle.State_Selected != 0:
            itemOption.state = itemOption.state & ~QStyle.State_Selected

        # Draw the rectangle
        pen = painter.pen()
        pen.setJoinStyle(Qt.RoundJoin)
        painter.setPen(pen)
        QGraphicsRectItem.paint(self, painter, itemOption, widget)

        # Draw text over the rectangle
        painter.setFont(self.__font)
        painter.setPen(QPen(QColor(255, 255, 255)))
        painter.drawText(self.__node.posX - self.__node.width / 2.0,
                         self.__node.posY - self.__node.height / 2.0,
                         self.__node.width, self.__node.height,
                         Qt.AlignCenter,
                         self.__node.label.replace('\\n', '\n'))

        if self.__outside:
            pixmap = getPixmap("nonprojectentrydgm.png")
            pixmapPosX = self.__node.posX - self.__node.width / 2.0 + 2
            pixmapPosY = self.__node.posY + self.__node.height / 2.0 - \
                pixmap.height() - 2
            painter.setRenderHint(QPainter.SmoothPixmapTransform)
            painter.drawPixmap(pixmapPosX, pixmapPosY, pixmap)

    def mouseDoubleClickEvent(self, _):
        """Open the clicked file if it could be opened"""
        if self.__lineNumber == 0:
            return
        if not os.path.isabs(self.__fileName):
            return

        GlobalData().mainWindow.openFile(self.__fileName,
                                         self.__lineNumber)


class ProfileGraphViewer(QWidget):

    """Profiling results as a graph"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, scriptName, params, reportTime,
                 dataFile, stats, parent=None):
        QWidget.__init__(self, parent)

        self.__dataFile = dataFile
        self.__script = scriptName
        self.__reportTime = reportTime
        self.__params = params
        self.__stats = stats

        project = GlobalData().project
        if project.isLoaded():
            self.__projectPrefix = os.path.dirname(project.fileName)
        else:
            self.__projectPrefix = os.path.dirname(scriptName)
        if not self.__projectPrefix.endswith(os.path.sep):
            self.__projectPrefix += os.path.sep

        self.__createLayout()
        self.__getDiagramLayout()

        self.__viewer.setScene(self.__scene)

    def setFocus(self):
        """Sets the focus properly"""
        self.__viewer.setFocus()

    def __isOutsideItem(self, fileName):
        """Detects if the record should be shown as an outside one"""
        return not fileName.startswith(self.__projectPrefix)

    def __createLayout(self):
        """Creates the widget layout"""
        totalCalls = self.__stats.total_calls
        # The calls were not induced via recursion
        totalPrimitiveCalls = self.__stats.prim_calls
        totalTime = self.__stats.total_tt

        txt = "<b>Script:</b> " + self.__script + " " + \
              self.__params['arguments'] + "<br/>" \
              "<b>Run at:</b> " + self.__reportTime + "<br/>" + \
              str(totalCalls) + " function calls (" + \
              str(totalPrimitiveCalls) + " primitive calls) in " + \
              FLOAT_FORMAT % totalTime + " CPU seconds"
        summary = HeaderFitLabel(self)
        summary.setText(txt)
        summary.setToolTip(txt)
        summary.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Fixed)
        summary.setMinimumWidth(10)

        self.__scene = QGraphicsScene()
        self.__viewer = DiagramWidget()
        self.__viewer.sigEscapePressed.connect(self.__onESC)

        vLayout = QVBoxLayout()
        vLayout.setContentsMargins(0, 0, 0, 0)
        vLayout.setSpacing(0)
        vLayout.addWidget(summary)
        vLayout.addWidget(self.__viewer)

        self.setLayout(vLayout)

    @staticmethod
    def __getDotFont(parts):
        """Provides a QFont object if a font spec is found"""
        for part in parts:
            if 'fontname=' in part:
                fontName = part.replace('fontname=', '')
                fontName = fontName.replace('[', '')
                fontName = fontName.replace(']', '')
                fontName = fontName.replace(',', '')
                return QFont(fontName)
        return None

    def __postprocessFullDotSpec(self, dotSpec):
        """Removes the arrow size, extracts tooltips, extracts font info"""
        nodeFont = None
        edgeFont = None
        tooltips = {}
        processed = []

        for line in dotSpec.splitlines():
            parts = line.split()
            lineModified = False
            if parts:
                if parts[0] == 'node':
                    # need to extract the fontname
                    nodeFont = self.__getDotFont(parts)
                elif parts[0] == 'edge':
                    # need to extract the fontname
                    edgeFont = self.__getDotFont(parts)
                elif parts[0].isdigit():
                    if parts[1] == '->':
                        # certain edge spec: replace arrowsize and font size
                        for index, part in enumerate(parts):
                            if part.startswith('[arrowsize='):
                                modified = parts[:]
                                modified[index] = '[arrowsize="0.0",'
                                processed.append(' '.join(modified))
                            elif part.startswith('fontsize='):
                                size = float(part.split('"')[1])
                                if edgeFont:
                                    edgeFont.setPointSize(size)
                        lineModified = True
                    elif parts[1].startswith('['):
                        # certain node spec: pick the tooltip and font size
                        lineno = None
                        for part in parts:
                            if part.startswith('tooltip='):
                                nodePath = part.split('"')[1]
                                pathLine = nodePath + ':' + str(lineno)
                                tooltips[int(parts[0])] = pathLine
                            elif part.startswith('fontsize='):
                                size = float(part.split('"')[1])
                                if nodeFont:
                                    nodeFont.setPointSize(size)
                            elif part.startswith('label='):
                                try:
                                    lineno = int(part.split(':')[1])
                                except:
                                    pass
            if not lineModified:
                processed.append(line)

        return '\n'.join(processed), tooltips, nodeFont, edgeFont


    def __rungprof2dot(self):
        """Runs gprof2dot which produces a full dot spec"""
        nodeLimit = Settings().getProfilerSettings().nodeLimit
        edgeLimit = Settings().getProfilerSettings().edgeLimit
        with io.StringIO() as buf:
            gprofParser = gprof2dot.PstatsParser(self.__dataFile)
            profileData = gprofParser.parse()
            profileData.prune(nodeLimit / 100.0, edgeLimit / 100.0, False, False)

            dot = gprof2dot.DotWriter(buf)
            dot.strip = False
            dot.wrap = False
            dot.graph(profileData, gprof2dot.TEMPERATURE_COLORMAP)

            output = buf.getvalue()
        return self.__postprocessFullDotSpec(output)

    def __getDiagramLayout(self):
        """Runs external tools to get the diagram layout"""
        fullDotSpec, tooltips, nodeFont, edgeFont = self.__rungprof2dot()

        dotProc = Popen(["dot", "-Tplain"],
                        stdin=PIPE, stdout=PIPE, bufsize=1)
        graphDescr = dotProc.communicate(
            fullDotSpec.encode('utf-8'))[0].decode('utf-8')

        graph = getGraphFromPlainDotData(graphDescr)
        graph.normalize(self.physicalDpiX(), self.physicalDpiY())

        self.__scene.clear()
        self.__scene.setSceneRect(0, 0, graph.width, graph.height)

        for edge in graph.edges:
            self.__scene.addItem(FuncConnection(edge))
            if edge.label != "":
                self.__scene.addItem(FuncConnectionLabel(edge, edgeFont))

        for node in graph.nodes:
            fileName = ""
            lineNumber = 0

            try:
                nodeNameAsInt = int(node.name)
                if nodeNameAsInt in tooltips:
                    parts = tooltips[nodeNameAsInt].rsplit(':', 1)
                    fileName = parts[0]
                    if parts[1].isdigit():
                        lineNumber = int(parts[1])
            except:
                pass

            self.__scene.addItem(Function(node, fileName, lineNumber,
                                          self.__isOutsideItem(fileName),
                                          nodeFont))

    def __onESC(self):
        """Triggered when ESC is clicked"""
        self.sigEscapePressed.emit()

    def onCopy(self):
        """Copies the diagram to the exchange buffer"""
        self.__viewer.onCopy()

    def onSaveAs(self, fileName):
        """Saves the diagram to a file"""
        self.__viewer.onSaveAs(fileName)

    def zoomIn(self):
        """Triggered on the 'zoom in' button"""
        self.__viewer.zoomIn()

    def zoomOut(self):
        """Triggered on the 'zoom out' button"""
        self.__viewer.zoomOut()

    def resetZoom(self):
        """Triggered on the 'zoom reset' button"""
        self.__viewer.resetZoom()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""UI support for detected outside changes of files"""


from utils.globals import GlobalData
from .qt import (Qt, QEventLoop, pyqtSignal, QColor, QFontMetrics, QSizePolicy,
                 QFrame, QLabel, QPushButton, QApplication, QGridLayout)


class OutsideChangeWidget(QFrame):

    """Frameless dialogue to deal with outside changes"""

    sigReloadRequest = pyqtSignal()
    reloadAllNonModifiedRequest = pyqtSignal()

    def __init__(self, parent):
        QFrame.__init__(self, parent)

        # Make the frame nice looking
        panelColor = QColor(170, 17, 17)
        palette = self.palette()
        palette.setColor(self.backgroundRole(), panelColor)
        self.setPalette(palette)

        self.setFrameShape(QFrame.StyledPanel)
        self.setLineWidth(2)
        self.setAutoFillBackground(True)

        # Keep pylint happy
        self.__messageLabel = None
        self.__leaveAsIsButton = None
        self.__reloadButton = None
        self.__reloadAllNonChangedButton = None

        self.__markers = []
        self.__createLayout()

        for item in self.__markers:
            item.hide()

    def __createLayout(self):
        """Creates the widget layout"""
        self.__messageLabel = QLabel("This file has been modified "
                                     "outside of codimension. What "
                                     "would you like to do?")
        self.__messageLabel.setWordWrap(True)
        self.__messageLabel.setAlignment(Qt.AlignHCenter)
        palette = self.__messageLabel.palette()
        palette.setColor(self.foregroundRole(), QColor(255, 255, 255, 255))
        self.__messageLabel.setPalette(palette)

        self.__leaveAsIsButton = QPushButton("Continue editing", self)
        self.__leaveAsIsButton.setToolTip("ESC")
        self.__leaveAsIsButton.clicked.connect(self.hide)

        self.__reloadButton = QPushButton(self)
        self.__reloadButton.clicked.connect(self.__reload)

        txt = "Reload all non-modified buffers"
        self.__reloadAllNonChangedButton = QPushButton(txt, self)
        self.__reloadAllNonChangedButton.clicked.connect(
            self.__reloadAllNonModified)

        # This will prevent the buttons growing wider than necessary
        fontMetrics = QFontMetrics(self.__reloadAllNonChangedButton.font())
        buttonWidth = fontMetrics.width(txt) + 20
        self.__reloadAllNonChangedButton.setFixedWidth(buttonWidth)

        gridLayout = QGridLayout(self)
        gridLayout.setContentsMargins(3, 3, 3, 3)

        gridLayout.addWidget(self.__messageLabel, 0, 0, 1, 1)
        gridLayout.addWidget(self.__leaveAsIsButton, 0, 1, 1, 1)
        gridLayout.addWidget(self.__reloadButton, 1, 1, 1, 1)
        gridLayout.addWidget(self.__reloadAllNonChangedButton, 2, 1, 1, 1)

        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.move(5, 5)


        self.__markers.append(QFrame(self.parent()))
        self.__markers.append(QFrame(self.parent()))
        self.__markers.append(QFrame(self.parent()))
        self.__markers.append(QFrame(self.parent()))

        markerColor = QColor(224, 0, 0, 255)
        for item in self.__markers:
            pal = item.palette()
            pal.setColor(item.backgroundRole(), markerColor)
            item.setPalette(pal)
            item.setFrameShape(QFrame.StyledPanel)
            item.setLineWidth(1)
            item.setAutoFillBackground(True)
            item.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

    def resize(self):
        """Resizes the dialogue to match the editor size"""
        vscroll = self.parent().verticalScrollBar()
        if vscroll.isVisible():
            scrollWidth = vscroll.width()
        else:
            scrollWidth = 0

        hscroll = self.parent().horizontalScrollBar()
        if hscroll.isVisible():
            scrollHeight = hscroll.height()
        else:
            scrollHeight = 0

        # Dialogue
        width = self.parent().width()
        height = self.parent().height()
        widgetWidth = width - scrollWidth - 10 - 1

        self.setFixedWidth(widgetWidth)

        # Marker
        self.__markers[0].move(1, 1)
        self.__markers[0].setFixedWidth(width - scrollWidth - 4)
        self.__markers[0].setFixedHeight(3)

        self.__markers[1].move(width - scrollWidth - 5, 1)
        self.__markers[1].setFixedWidth(3)
        self.__markers[1].setFixedHeight(height - scrollHeight - 4)

        self.__markers[2].move(1, height - scrollHeight - 5)
        self.__markers[2].setFixedWidth(width - scrollWidth - 4)
        self.__markers[2].setFixedHeight(3)

        self.__markers[3].move(1, 1)
        self.__markers[3].setFixedWidth(3)
        self.__markers[3].setFixedHeight(height - scrollHeight - 4)

    def showChoice(self, modified, allEnabled):
        """Brings up the panel with the correct text and buttons"""
        if modified:
            self.__reloadButton.setText("Reload losing changes")
        else:
            self.__reloadButton.setText("Reload")
        self.__reloadAllNonChangedButton.setEnabled(allEnabled)

        QApplication.processEvents(QEventLoop.ExcludeUserInputEvents)

        self.resize()
        self.show()
        for item in self.__markers:
            item.show()
        self.__leaveAsIsButton.setFocus()
        self.parent().setReadOnly(True)

    def setFocus(self):
        """Passes the focus to the default button"""
        self.__leaveAsIsButton.setFocus()

    def keyPressEvent(self, event):
        """Handles the key press events"""
        if event.key() == Qt.Key_Escape:
            editorsManager = GlobalData().mainWindow.editorsManager()
            activeWindow = editorsManager.currentWidget()
            if activeWindow:
                activeWindow.setFocus()
            event.accept()
            self.hide()

    def hide(self):
        """Handles the hiding of the panel and markers"""
        for item in self.__markers:
            item.hide()
        QFrame.hide(self)
        self.parent().setReadOnly(False)
        self.parent().setFocus()

    def __reload(self):
        """Reloads the file from the disk"""
        self.sigReloadRequest.emit()

    def __reloadAllNonModified(self):
        """Reloads all the non-modified buffers"""
        self.reloadAllNonModifiedRequest.emit()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Base class for all the main window tab widgets"""

import uuid


class MainWindowTabWidgetBase():

    """Base class for all the main window tab widgets"""

    Unknown = -1
    PlainTextEditor = 0     # Text editor + may be a graphics widget
    PictureViewer = 1
    HTMLViewer = 2
    GeneratedDiagram = 3
    ProfileViewer = 4
    VCSAnnotateViewer = 5
    DiffViewer = 6
    MDViewer = 7

    def __init__(self):
        self.__uuid = str(uuid.uuid1())
        self.__tooltip = ""

    def isModified(self):
        """Tells if the file is modifed"""
        raise Exception("isModified() is not implemented")

    def getRWMode(self):
        """Tells the read/write mode"""
        return None

    def getType(self):
        """Tells the widget type"""
        raise Exception("getType() is not implemented")

    def getLanguage(self):
        """Tells the content language"""
        raise Exception("getLanguage() is not implemented")

    def getFileName(self):
        """Tells what file name of the widget"""
        return None

    def setFileName(self, path):
        """Sets the file name"""
        raise Exception("setFilename() is not implemented")

    def getEol(self):
        """Tells the EOL style"""
        return None

    def getLine(self):
        """Tells the cursor line"""
        return None

    def getPos(self):
        """Tells the cursor column"""
        return None

    def getEncoding(self):
        """Tells the content encoding"""
        return None

    def setEncoding(self, newEncoding):
        """Sets the encoding for the text document"""
        raise Exception("setEncoding() is not implemented")

    def getShortName(self):
        """Tells the display name"""
        raise Exception("getShortName() is not implemented")

    def setShortName(self, name):
        """Sets the display name"""
        raise Exception("setShortName() is not implemented")

    def getUUID(self):
        """Provides the widget unique ID"""
        return self.__uuid

    def isDiskFileModified(self):
        """Return True if the loaded file is modified"""
        return False

    def doesFileExist(self):
        """Returns True if the loaded file still exists"""
        return True

    def setReloadDialogShown(self, value=True):
        """Memorizes if the reloading dialogue has already been displayed"""

    def getReloadDialogShown(self):
        """Tells if the reload dialog has already been shown"""
        return True

    def showOutsideChangesBar(self, allEnabled):
        """Shows the outside changes bar"""

    def reload(self):
        """Reloads the widget content from the file"""

    def setTooltip(self, txt):
        """Saves the tab tooltip"""
        self.__tooltip = txt

    def getTooltip(self):
        """Returns the saved tooltip"""
        return self.__tooltip

    def setDebugMode(self, mode, isProjectFile):
        """Switches the widget to debug mode and back"""

    def getVCSStatus(self):
        """Provides the content VCS status"""
        return None
